#!/usr/bin/env python3
"""SC-Controller - Menu Editor

Edits .menu files and menus stored in profile.
"""

import logging
import os
import traceback

from gi.repository import GObject, Gtk

from scc.actions import Action, NoAction
from scc.gui.action_editor import ActionEditor
from scc.gui.dwsnc import headerbar
from scc.gui.editor import Editor
from scc.gui.icon_chooser import IconChooser
from scc.menu_data import MenuData, MenuItem, Separator, Submenu
from scc.osd.menu import MenuIcon
from scc.osd.menu_generators import GameListMenuGenerator, ProfileListMenuGenerator, RecentListMenuGenerator
from scc.parser import TalkingActionParser
from scc.paths import get_default_menus_path, get_menus_path
from scc.profile import Encoder
from scc.tools import _, find_icon

log = logging.getLogger("MenuEditor")


class MenuEditor(Editor):
	GLADE = "menu_editor.glade"
	TYPE_INTERNAL = 1
	TYPE_GLOBAL = 2

	OPEN = set()  # Set of menus that are being edited.

	def __init__(self, app, callback):
		self.app = app
		self.next_auto_id = 1
		self.callback = callback
		self.selected_icon = None
		self.original_id = None
		self.original_type = MenuEditor.TYPE_INTERNAL
		Editor.install_error_css()
		self.setup_widgets()

	def setup_widgets(self):
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(self.app.gladepath, self.GLADE))
		lblItemIconName = self.builder.get_object("lblItemIconName")
		vbChangeItemIcon = self.builder.get_object("vbChangeItemIcon")
		self.window = self.builder.get_object("Dialog")
		self.menu_icon = MenuIcon(None, True)
		vbChangeItemIcon.remove(lblItemIconName)
		vbChangeItemIcon.pack_start(self.menu_icon, False, True, 0)
		vbChangeItemIcon.pack_start(lblItemIconName, True, True, 0)
		self.builder.connect_signals(self)
		headerbar(self.builder.get_object("header"))

	def allow_menus(self, allow_globals, allow_in_profile):
		"""Sets which type of menu should be selectable.
		By default, both are enabled.
		"""
		if not allow_globals:
			self.builder.get_object("rbInProfile").set_active(True)
			self.builder.get_object("rbGlobal").set_sensitive(False)
			self.builder.get_object("rbInProfile").set_sensitive(True)
		elif not allow_in_profile:
			self.builder.get_object("rbGlobal").set_active(True)
			self.builder.get_object("rbGlobal").set_sensitive(True)
			self.builder.get_object("rbInProfile").set_sensitive(False)
		else:
			self.builder.get_object("rbGlobal").set_sensitive(True)
			self.builder.get_object("rbInProfile").set_sensitive(True)

	def on_action_chosen(self, id, a, mark_changed=True):
		model = self.builder.get_object("tvItems").get_model()
		for i in model:
			item = i[0].item
			if item.id == id:
				if isinstance(item, Separator):
					item.label = a.get_name()
				elif isinstance(item, Submenu):
					i[0].item = item = Submenu(a.get_current_page().get_selected_menu(), a.get_name())
					item.icon = self.selected_icon
				elif isinstance(item, RecentListMenuGenerator):
					i[0].item = item = RecentListMenuGenerator(rows=a.get_current_page().get_row_count())
				elif isinstance(item, MenuItem):
					item.action = a
					item.label = item.action.describe(Action.AC_OSD)
					item.icon = self.selected_icon
				else:
					raise TypeError("Edited %s" % (item.__class__.__name__))
				i[1] = item.describe()
				break

	def on_btSave_clicked(self, *a):
		"""Handler for Save button"""
		self._remove_original()
		if self.builder.get_object("rbInProfile").get_active():
			self._save_to_profile(self.builder.get_object("entName").get_text())
		else:
			self._save_to_file(self.builder.get_object("entName").get_text())
		self.close()

	def on_tvItems_cursor_changed(self, *a):
		"""Handles moving cursor in Item List.
		Basically just sets Edit Item and Remove Item buttons sensitivity.
		"""
		tvItems = self.builder.get_object("tvItems")
		btEdit = self.builder.get_object("btEdit")
		btRemoveItem = self.builder.get_object("btRemoveItem")

		model, iter = tvItems.get_selection().get_selected()
		if iter is None:
			btRemoveItem.set_sensitive(False)
			btEdit.set_sensitive(False)
		else:
			btRemoveItem.set_sensitive(True)
			o = model.get_value(iter, 0)
			if isinstance(o.item, (MenuItem, RecentListMenuGenerator)):
				btEdit.set_sensitive(True)
			else:
				btEdit.set_sensitive(False)

	def btEdit_clicked_cb(self, *a):
		"""Handler for "Edit Item" button"""
		tvItems = self.builder.get_object("tvItems")
		model, iter = tvItems.get_selection().get_selected()
		item = model.get_value(iter, 0).item
		self.selected_icon = None
		# Setup editor
		e = ActionEditor(self.app, self.on_action_chosen)
		if isinstance(item, Separator):
			e.set_title(_("Edit Separator"))
			e.hide_editor()
			e.set_menu_item(item, _("Separator Name"))
		elif isinstance(item, Submenu):
			e.set_title(_("Edit Submenu"))
			e.hide_action_str()
			e.hide_clear()
			(
				e.force_page(e.load_component("menu_only"), True)
				.allow_menus(True, False)
				.set_selected_menu(item.filename)
			)
			e.set_menu_item(item, _("Menu Label"))
			self.selected_icon = item.icon
			self.setup_menu_icon(e)
			self.update_menu_icon()
		elif isinstance(item, MenuItem):
			e = ActionEditor(self.app, self.on_action_chosen)
			e.set_title(_("Edit Menu Action"))
			e.set_input(item.id, item.action, mode=Action.AC_MENU)
			self.selected_icon = item.icon
			self.setup_menu_icon(e)
			self.update_menu_icon()
		elif isinstance(item, RecentListMenuGenerator):
			e.set_title(_("Edit Recent List"))
			e.hide_action_str()
			e.hide_clear()
			e.hide_name()
			(e.force_page(e.load_component("recent_list"), True).set_row_count(item.rows))
			e.set_menu_item(item)
		else:
			# Cannot edit this
			return
		# Display editor
		e.show(self.window)

	def _add_menuitem(self, item):
		"""Adds MenuItem or MenuGenerator object"""
		tvItems = self.builder.get_object("tvItems")
		model = tvItems.get_model()
		o = GObject.GObject()
		if not item.id:
			item.id = "_auto_id_%s" % (self.next_auto_id,)
			self.next_auto_id += 1
		o.item = item
		iter = model.append((o, o.item.describe()))
		tvItems.get_selection().select_iter(iter)
		self.on_tvItems_cursor_changed()

	def on_btAddItem_clicked(self, *a):
		"""Handler for "Add Action" button and menu item"""
		item = MenuItem(None, NoAction().describe(Action.AC_OSD), NoAction())
		self._add_menuitem(item)
		self.btEdit_clicked_cb()

	def on_mnuAddSeparator_clicked(self, *a):
		"""Handler for "Add Separator" menu item"""
		self._add_menuitem(Separator())

	def on_mnuAddSubmenu_clicked(self, *a):
		"""Handler for "Add Separator" menu item"""
		self._add_menuitem(Submenu(""))

	def on_mnuAddProfList_clicked(self, *a):
		"""Handler for "Add List of All Profiles" menu item"""
		self._add_menuitem(ProfileListMenuGenerator())

	def on_mnuAddRecentList_clicked(self, *a):
		"""Handler for "Add List of Recent Profiles" menu item"""
		self._add_menuitem(RecentListMenuGenerator())

	def on_mnuAddGamesList_activate(self, *a):
		"""Handler for "Add List of Games" menu item"""
		self._add_menuitem(GameListMenuGenerator())

	def on_btRemoveItem_clicked(self, *a):
		"""Handler for "Delete Item" button"""
		tvItems = self.builder.get_object("tvItems")
		model, iter = tvItems.get_selection().get_selected()
		if iter is not None:
			model.remove(iter)
		self.on_tvItems_cursor_changed()

	def on_entName_changed(self, *a):
		id = self.builder.get_object("entName").get_text()
		if len(id.strip()) == 0:
			self._bad_id_no_id()
			return
		if "." in id or "/" in id:
			self._bad_id_chars()
			return
		if self.builder.get_object("rbInProfile").get_active():
			# Menu stored in profile
			if id != self.original_id and id in self.app.current.menus:
				self._bad_id_duplicate()
				return
		# Menu stored as file
		elif id != self.original_id:
			path = os.path.join(get_menus_path(), "%s.menu" % (id,))
			if os.path.exists(path):
				self._bad_id_duplicate()
				return
		self._good_id()
		return

	def _good_id(self, *a):
		self.builder.get_object("rvInvalidID").set_reveal_child(False)
		self.builder.get_object("btSave").set_sensitive(True)

	def _bad_id_no_id(self, *a):
		self.builder.get_object("btSave").set_sensitive(False)

	def _bad_id_duplicate(self, *a):
		self.builder.get_object("lblNope").set_label(_("Invalid Menu ID: Menu with same ID already exists."))
		self.builder.get_object("rvInvalidID").set_reveal_child(True)
		self.builder.get_object("btSave").set_sensitive(False)

	def _bad_id_chars(self, *a):
		self.builder.get_object("lblNope").set_label(_("Invalid Menu ID: Please, don't use dots (.) or slashes (/)."))
		self.builder.get_object("rvInvalidID").set_reveal_child(True)
		self.builder.get_object("btSave").set_sensitive(False)

	def set_new_menu(self):
		"""Setups editor for creating new menu.
		"""
		self.set_title(_("New Menu"))
		rbInProfile = self.builder.get_object("rbInProfile")
		entName = self.builder.get_object("entName")

		rbInProfile.set_active(True)
		self.original_id = None
		self.original_type = MenuEditor.TYPE_INTERNAL
		entName.set_text("")

	@staticmethod
	def menu_is_global(id):
		return "." in id

	def set_menu(self, id):
		"""Setups editor for menu with specified ID.
		ID may be id of menu in profile or, if it contains ".", filename.
		"""
		self.set_title(_("Menu Editor"))
		rbGlobal = self.builder.get_object("rbGlobal")
		rbInProfile = self.builder.get_object("rbInProfile")
		entName = self.builder.get_object("entName")

		MenuEditor.OPEN.add(id)
		if MenuEditor.menu_is_global(id):
			id = id.split(".")[0]
			rbGlobal.set_active(True)
			self.original_type = MenuEditor.TYPE_GLOBAL
			items = self._load_items_from_file(id)
		else:
			rbInProfile.set_active(True)
			self.original_type = MenuEditor.TYPE_INTERNAL
			items = self._load_items_from_profile(id)
		self.original_id = id
		entName.set_text(id)

		model = self.builder.get_object("tvItems").get_model()
		model.clear()
		if items is None:
			self.set_new_menu()
		else:
			for i in items:
				self._add_menuitem(i)

	def on_Dialog_delete_event(self, *a):
		try:
			if self.original_type == MenuEditor.TYPE_GLOBAL:
				MenuEditor.OPEN.remove(self.original_id + ".menu")
			else:
				MenuEditor.OPEN.remove(self.original_id)
		except KeyError:
			pass
		return False

	def close(self, *a):
		self.on_Dialog_delete_event()
		Editor.close(self)

	def _load_items_from_file(self, id):
		for p in (get_menus_path(), get_default_menus_path()):
			path = os.path.join(p, "%s.menu" % (id,))
			if os.path.exists(path):
				return MenuData.from_file(path, TalkingActionParser())
		# Menu file not found
		return None

	def _load_items_from_profile(self, id):
		try:
			return self.app.current.menus[id]
		except KeyError:
			# Menu not found
			return None

	def _remove_original(self):
		if self.original_id is None:
			# Created new menu
			pass
		elif self.original_type == MenuEditor.TYPE_INTERNAL:
			try:
				del self.app.current.menus[self.original_id]
			except:
				pass
		elif self.original_type == MenuEditor.TYPE_GLOBAL:
			try:
				path = os.path.join(get_menus_path(), "%s.menu" % (self.original_id,))
				log.debug("Removing %s", path)
				os.unlink(path)
			except:
				pass

	def _generate_menudata(self):
		"""Generates MenuData instance from items in list
		"""
		model = self.builder.get_object("tvItems").get_model()
		data = MenuData(*[i[0].item for i in model])
		i = 1
		for item in data:
			item.id = "item%s" % (i,)
			i += 1
		return data

	def _save_to_profile(self, id):
		"""Stores menu in loaded profile. Doesn't actually save anything, that's
		for main app window thing.
		"""
		self.app.current.menus[id] = self._generate_menudata()
		log.debug("Stored menu ID %s", id)
		self.app.on_profile_modified()
		if self.callback:
			self.callback(id)

	def _save_to_file(self, id):
		"""Stores menu in json file
		"""
		id = "%s.menu" % (id,)
		path = os.path.join(get_menus_path(), id)
		data = self._generate_menudata()
		jstr = Encoder(sort_keys=True, indent=4).encode(data)
		open(path, "w").write(jstr)
		log.debug("Wrote menu file %s", path)
		if self.callback:
			self.callback(id)

	def setup_menu_icon(self, editor):
		container = self.builder.get_object("menu_icon")
		editor.add_widget(_("Icon"), container)

	def update_menu_icon(self):
		lblItemIconName = self.builder.get_object("lblItemIconName")
		if self.selected_icon is None:
			lblItemIconName.set_label(_("(no icon)"))
			self.menu_icon.set_visible(False)
		else:
			lblItemIconName.set_label(self.selected_icon)
			try:
				filename, trash = find_icon(self.selected_icon)
				self.menu_icon.set_filename(filename)
				self.menu_icon.set_visible(True)
			except Exception as e:
				log.error(e)
				log.error(traceback.format_exc())
				self.menu_icon.set_visible(False)

	def on_icon_choosen(self, name):
		self.selected_icon = name
		self.update_menu_icon()

	def on_btChangeItemIcon_clicked(self, *a):
		c = IconChooser(self.app, self.on_icon_choosen)
		c.show(self.window)

	def on_btClearItemIcon_clicked(self, *a):
		self.selected_icon = None
		self.update_menu_icon()



"""hidparse - just enough code to parse HID report from hidraw descriptor.

Based on
  - Pythonic binding for linux's hidraw ioctls
    (https://github.com/vpelletier/python-hidraw)
  - Winfred Lu's rd-parse.py
    (http://winfred-lu.blogspot.sk/2014/02/usb-hid-report-descriptor-parser-in.html)

Licensed under GPL 2.0
"""

import ctypes
import fcntl
import struct
from enum import IntEnum

import ioctl_opt

from scc.lib.hidparse_data import (
	Collection,
	GenericDesktopPage,
	GlobalItem,
	HidSensorProperty,
	ItemBase,
	ItemLength,
	ItemType,
	LightSensor,
	LocalItem,
	MainItem,
	ModifierI2a,
	MotionSensor,
	OrientationSensor,
	SensorDataField,
	SensorEvent,
	SensorPage,
	SensorSelector,
	Unit,
	UnitType,
	UsagePage,
	page_to_enum,
)

# hid.h
_HID_MAX_DESCRIPTOR_SIZE = 4096

AXES = [
	GenericDesktopPage.X,
	GenericDesktopPage.Y,
	GenericDesktopPage.Z,
	GenericDesktopPage.Rx,
	GenericDesktopPage.Ry,
	GenericDesktopPage.Rz,
]


# hidraw.h
class _hidraw_report_descriptor(ctypes.Structure):
	_fields_ = [
		("size", ctypes.c_uint),
		("value", ctypes.c_ubyte * _HID_MAX_DESCRIPTOR_SIZE),
	]


class _hidraw_devinfo(ctypes.Structure):
	_fields_ = [
		("bustype", ctypes.c_uint),
		("vendor", ctypes.c_short),
		("product", ctypes.c_short),
	]


class BusType(IntEnum):
	USB = 0x03
	HIL = 0x04
	BLUETOOTH = 0x05
	VIRTUAL = 0x06


class ReservedItem:
	_CACHE = {}

	def __init__(self, value):
		self.value = value

	def __repr__(self) -> str:
		return "<Reserved ID 0x%x>" % (self.value,)

	__str__ = __repr__

	def __new__(cls, value):
		if value not in cls._CACHE:
			cls._CACHE[value] = object.__new__(cls)
		return cls._CACHE[value]


def enum_or_reserved(enum, value):
	try:
		return enum(value)
	except ValueError:
		return ReservedItem(value)


_HIDIOCGRDESCSIZE = ioctl_opt.IOR(ord("H"), 0x01, ctypes.c_int)
_HIDIOCGRDESC = ioctl_opt.IOR(ord("H"), 0x02, _hidraw_report_descriptor)
_HIDIOCGRAWINFO = ioctl_opt.IOR(ord("H"), 0x03, _hidraw_devinfo)
# _HIDIOCGFEATURE =	lambda len : ioctl_opt.IOWR(ord('H'), 0x07, len)
_HIDIOCGFEATURE = lambda len: ioctl_opt.IOC(ioctl_opt.IOC_WRITE | ioctl_opt.IOC_READ, ord("H"), 0x07, len)


def _ioctl(devfile, func, arg, mutate_flag=False):
	result = fcntl.ioctl(devfile, func, arg, mutate_flag)
	if result < 0:
		raise OSError(result)


def get_device_info(devfile):
	"""Return tuple of (bustype, vendor_id, product_id), where bustype is instance of BusType enum."""
	devinfo = _hidraw_devinfo()
	_ioctl(devfile, _HIDIOCGRAWINFO, devinfo, True)
	return (BusType(devinfo.bustype), devinfo.vendor, devinfo.product)


def get_raw_report_descriptor(devfile):
	"""Return raw HID report descriptor as list of bytes."""
	descriptor = _hidraw_report_descriptor()
	size = ctypes.c_uint()
	_ioctl(devfile, _HIDIOCGRDESCSIZE, size, True)
	descriptor.size = size
	_ioctl(devfile, _HIDIOCGRDESC, descriptor, True)
	return descriptor.value[: size.value]


# Convert items to unsigned char, short, or int
def _it2u(it):
	if len(it) == 2:  # unsigned char
		n = it[1]
	elif len(it) == 3:  # unsigned short
		n = int(f"{it[2]:02x}{it[1]:02x}", 16)
	elif len(it) == 5:  # unsigned int
		n = int(f"{it[4]:02x}{it[3]:02x}{it[2]:02x}{it[1]:02x}", 16)
	else:
		n = 0
	return n


# Convert items to signed char, short, or int
def _it2s(it):
	if len(it) == 2:  # signed char
		n = it[1]
		if n & 0x80:
			n -= 0x100
	elif len(it) == 3:  # signed short
		n = int(f"{it[2]:02x}{it[1]:02x}", 16)
		if n & 0x8000:
			n -= 0x10000
	elif len(it) == 5:  # signed int
		n = int(f"{it[4]:02x}{it[3]:02x}{it[2]:02x}{it[1]:02x}", 16)
		if n & 0x80000000:
			n -= 0x100000000
	else:
		n = 0
	return n


def parse_item(it, page):
	if it[0] != 0xFE:  # normal item
		itag = it[0] & 0xF0
		itype = it[0] & 0xC
		isize = it[0] & 0x3
	else:  # long item
		isize = it[1]
		itag = it[3] * 256 + it[2]
		raise ValueError("Not implemented: long item!!")

	if itype == 0x00:  # main items
		item = enum_or_reserved(MainItem, itag)
		if item == MainItem.Collection:
			col_type = enum_or_reserved(Collection, it[1])
			return item, col_type
		if item in (MainItem.Input, MainItem.Output, MainItem.Feature):
			return (
				item,
				ItemType.Constant if it[1] & 0x1 else ItemType.Data,
				ItemLength.Variable if it[1] & 0x2 else ItemLength.Array,
				ItemBase.Relative if it[1] & 0x4 else ItemBase.Absolute,
			)
		# EndCollection or reserved
		return (item,)
	if itype == 0x04:  # global items
		item = enum_or_reserved(GlobalItem, itag)
		if item == GlobalItem.UsagePage:
			page = enum_or_reserved(UsagePage, _it2u(it))
			return item, page
		if item == GlobalItem.UnitExponent:
			# exponent
			value = it[1] if it[1] < 8 else it[1] - 0x10
			return item, value
		if item == GlobalItem.Unit:
			if it[1] == 0:
				return item, UnitType.NoUnit, None
			nibble = it[1] & 0x0F
			unit_type = enum_or_reserved(UnitType, nibble)
			if it[1] & 0xF0 != 0:
				return item, unit_type, Unit.Length
			if len(it) > 2:
				if it[2] & 0x0F:
					return item, unit_type, Unit.Mass
				if it[2] & 0xF0:
					return item, unit_type, Unit.Item
			if len(it) > 3:
				if it[3] & 0x0F:
					return item, unit_type, Unit.Temperature
				if it[3] & 0xF0:
					return item, unit_type, Unit.Current
			if len(it) > 4:
				if it[4] & 0x0F:
					return item, unit_type, Unit.LuminousIntensity
			return item, unit_type, ReservedItem(it[1])
		if item in (GlobalItem.LogicalMaximum, GlobalItem.PhysicalMaximum):
			# unsigned values
			return item, _it2u(it)
		if item in (GlobalItem.LogicalMinimum, GlobalItem.PhysicalMinimum, GlobalItem.ReportSize):
			# signed values
			return item, _it2s(it)
		if item in (GlobalItem.ReportID, GlobalItem.ReportCount):
			return item, it[1]
		return item
	if itype == 0x08:  # local items
		item = enum_or_reserved(LocalItem, itag)
		if item == LocalItem.Usage:
			if page is SensorPage and isize == 2:  # sensor page & usage size is 2
				mdf = (it[2] & 0xF0) >> 4
				if it[2] & 0xF == 0x02:
					return (item, enum_or_reserved(ModifierI2a, mdf), enum_or_reserved(SensorEvent, it[1]))
				if it[2] & 0xF == 0x03:
					return (item, enum_or_reserved(ModifierI2a, mdf), enum_or_reserved(HidSensorProperty, it[1]))
				if it[2] & 0xF == 0x04:
					if it[1] & 0xF0 == 0x50:
						return (item, enum_or_reserved(ModifierI2a, mdf), enum_or_reserved(MotionSensor, it[1]))
					if it[1] & 0xF0 == 0x70 or it[1] & 0xF0 == 0x80:
						return (item, enum_or_reserved(ModifierI2a, mdf), enum_or_reserved(OrientationSensor, it[1]))
					if it[1] & 0xF0 == 0xD0:
						return (item, enum_or_reserved(ModifierI2a, mdf), enum_or_reserved(LightSensor, it[1]))
				elif it[2] & 0xF == 0x05:
					return (item, enum_or_reserved(ModifierI2a, mdf), enum_or_reserved(SensorDataField, it[1]))
				elif it[2] & 0xF == 0x08:
					return (item, enum_or_reserved(ModifierI2a, mdf), enum_or_reserved(SensorSelector, it[1]))
			elif isize == 3:  # 4 bytes (usage page: usage id)
				uid = it[2] * 256 + it[1]
				upg = it[4] * 256 + it[3]
				page = enum_or_reserved(UsagePage, upg)
				try:
					return item, page(uid)
				except ValueError:
					return item, uid
				return item, page
			else:
				try:
					return item, page(it[1])
				except ValueError:
					return item, it[1]
		elif item in (
			LocalItem.UsageMinimum,
			LocalItem.UsageMaximum,
			LocalItem.DesignatorMinimum,
			LocalItem.DesignatorMaximum,
			LocalItem.StringMinimum,
			LocalItem.StringMaximum,
		):
			return item, _it2s(it)
		else:
			return item
	else:
		return ReservedItem(itype)


def _split_hid_items(data):
	size = 0
	for i in range(len(data)):
		if size != 0:  # skip bytes for the previous item
			size -= 1
			continue
		size = data[i] & 0x3  # 3 means 4 bytes
		if size == 3:
			size = 4
		if i == 0xFE:  # long item
			size = data[i + 1]
		yield data[i : i + size + 1]


def parse_report_descriptor(data, flat_list=False):
	"""Parses HID report descriptor to list of elements.

	If flat_list is set to True, only one list is returned.
	Otherwise, each collection is stored in its own nested list.
	"""
	rv, page = [], GenericDesktopPage
	stack, col = [], rv
	for it in _split_hid_items(data):
		item = parse_item(it, page)
		if not flat_list and item[0] is MainItem.Collection:
			item = [item]
			col.append(item)
			stack.append(col)
			col = item
		elif not flat_list and item[0] is MainItem.EndCollection:
			col.append(item)
			col, stack = stack[0], stack[:-1]
		elif item[0] is GlobalItem.UsagePage:
			page = item[1]
			if item[1] in page_to_enum:
				page = page_to_enum[item[1]]
			else:
				page = GenericDesktopPage
			col.append(item)
		else:
			col.append(item)

	return rv


def get_report_descriptor(devfile, flat_list=False):
	"""Returns parsed HID report descriptor as list of elements.

	If flat_list is set to True, only one list is returned.
	Otherwise, each collection is stored in its own nested list.
	"""
	data = get_raw_report_descriptor(devfile)
	return parse_report_descriptor(data, flat_list)


class Parser:
	def __init__(self, code, offset, count, size):
		self.code = code
		self.value = 0
		self.offset = offset
		self.byte_offset = offset / 8
		self.bit_offset = offset % 8
		self.count = count
		self.len = count * size
		if self.len > 64:
			raise ValueError("Too many bytes in value: %i" % (self.len,))
		if self.len > 32:
			self.byte_len = 8
			self.fmt = "<Q"
		elif self.len > 16:
			self.byte_len = 4
			self.fmt = "<I"
		elif self.len > 8:
			self.byte_len = 2
			self.fmt = "<H"
		else:
			self.byte_len = 1
			self.fmt = "<B"
		self.additional_bits = offset % 8

	def decode(self, data):
		(self.value,) = struct.unpack(self.fmt, data[self.byte_offset : self.byte_offset + self.byte_len])
		self.value >>= self.additional_bits


HIDPARSE_TYPE_AXIS = 1
HIDPARSE_TYPE_BUTTONS = 2


class HIDButtonParser(Parser):
	TYPE = HIDPARSE_TYPE_BUTTONS

	def __repr__(self):
		return "<HID Buttons @%s len %s value %s>" % (self.offset, self.len, self.value)


class HIDAxisParser(Parser):
	TYPE = HIDPARSE_TYPE_AXIS

	def __repr__(self):
		return "<HID Axis @%s len %s value %s>" % (self.offset, self.len, self.value)


def make_parsers(data):
	size, count = 1, 0
	kind = None
	offset = 0
	parsers = []
	axis_id, buttons_id = 0, 0
	for x in parse_report_descriptor(data, True):
		# print(x)
		if x[0] == GlobalItem.ReportSize:
			size = x[1]
		elif x[0] == GlobalItem.ReportCount:
			count = x[1]
		elif x[0] == LocalItem.Usage:
			kind = x[1]
		elif x[0] == MainItem.Input:
			if x[1] == ItemType.Constant:
				pass
			elif x[1] == ItemType.Data:
				if kind in AXES:
					for i in range(count):
						parsers.append(HIDAxisParser(axis_id, offset + size * i, 1, size))
						axis_id += 1
				else:
					parsers.append(HIDButtonParser(buttons_id, offset, count, size))
					buttons_id += count
			offset += size * count
	size = offset / 8
	if offset % 8 > 0:
		size += 1
	return size, parsers



"""SC Controller - Scheduler

Centralized scheduler that should be used everywhere.
Runs in SCCDaemon's (single-threaded) mainloop. That means all callbacks are
also called on main thread.

Use schedule(delay, callback, *data) to register one-time task.
"""

import logging
import queue
import time

log = logging.getLogger("Scheduler")

# TODO: Maybe create actual thread for this? Use poler? Scrap everything and rewrite it in GO?


class Scheduler:
	def __init__(self):
		self._scheduled = queue.PriorityQueue()
		self._next = None
		self._now = time.time()

	def schedule(self, delay, callback, *data):
		"""Schedules one-time task to be executed no sooner than after 'delay' of
		seconds. Delay may be float number.
		'callback' is called as callback(*data).

		Returned Task instance can be used to cancel task once scheduled.
		"""
		task = Task(self._now + delay, callback, data)
		if self._next is None or task.time < self._next.time:
			if self._next:
				self._scheduled.put(self._next)
			self._next = task
		else:
			self._scheduled.put(task)
		return task

	def cancel_task(self, task):
		"""Returns True if task was sucessfully removed or False if task was
		already executed or not known at all.

		Note that this is slow as hell and completly thread-unsafe,
		so it _has_ to be called on main thread.
		"""
		if task == self._next:
			self._next = None if self._scheduled.empty() else self._scheduled.get()
			return True
		# Fun part: All tasks are removed from PriorityQueue
		# until correct is found. Then everything is put back
		tasks, found = [], False
		while not self._scheduled.empty():
			t = self._scheduled.get()
			if t == task:
				found = True
				break
			tasks.append(t)
		for t in tasks:
			self._scheduled.put(t)
		return found

	def run(self):
		self._now = time.time()
		while self._next and self._now >= self._next.time:
			callback, data = self._next.callback, self._next.data
			self._next = None if self._scheduled.empty() else self._scheduled.get()
			callback(*data)


class Task:
	def __init__(self, time, callback, data):
		self.time = time
		self.callback = callback
		self.data = data

	def cancel(self):
		"""Marks task as canceled, without actually removing it from scheduler"""
		self.callback = lambda *a, **b: False
		self.data = ()

	def __lt__(self, other):
		self.time < other.time



#!/usr/bin/env python3
"""SC-Controller - Action Editor - Recent List Component

Displays page that can edit settings for RecentListMenuGenerator
"""

import logging

from scc.gui.ae import AEComponent
from scc.tools import _

log = logging.getLogger("AE.SA")

__all__ = ["RecentListGenComponent"]


class RecentListGenComponent(AEComponent):
	GLADE = "ae/recent_list.glade"
	NAME = "recent_list"
	CTXS = 0
	PRIORITY = 0

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)

	def set_action(self, mode, action):
		pass

	def get_button_title(self):
		return _("Recent List")

	def handles(self, mode, action):
		"""Not visible by default"""
		return False

	def set_row_count(self, count):
		self.builder.get_object("sclNumOfProfiles").set_value(count)

	def get_row_count(self):
		return int(self.builder.get_object("sclNumOfProfiles").get_value())



"""SC-Controller - Action Editor - Trigger-as-button Component.

Assigns one or two emulated buttons to trigger
"""

import logging

from scc.actions import (
	Action,
	AxisAction,
	ButtonAction,
	HipfireAction,
	MouseAction,
	MultiAction,
	NoAction,
	TriggerAction,
)
from scc.constants import (
	HIPFIRE_EXCLUSIVE,
	HIPFIRE_NORMAL,
	HIPFIRE_SENSIBLE,
	TRIGGER_CLICK,
	TRIGGER_HALF,
	TRIGGER_MAX,
	TRIGGER_MIN,
)
from scc.gui.ae import AEComponent, describe_action
from scc.gui.binding_editor import BindingEditor
from scc.gui.simple_chooser import SimpleChooser
from scc.tools import _

log = logging.getLogger("AE.TriggerAB")

__all__ = ["TriggerComponent"]


class TriggerComponent(AEComponent, BindingEditor):
	GLADE = "ae/trigger.glade"
	NAME = "trigger"
	CTXS = Action.AC_TRIGGER

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		BindingEditor.__init__(self, app)
		self._recursing = False
		self.half = NoAction()
		self.full = NoAction()
		self.analog = NoAction()

	def handles(self, mode, action):
		if isinstance(action, NoAction):
			return True
		sucess, half, full, analog = TriggerComponent._split(action)
		return sucess

	@staticmethod
	def _split(action):
		"""Split passed action so it can be displayed in UI.

		Returns (sucess, half, full, analog), with three actions
		for each UI element.
		Note that each returned action may be TriggerAction.

		If passed action cannot be decoded,
		'sucess' element of tuple is set to False
		"""
		half, full, analog = NoAction(), NoAction(), NoAction()
		actions = action.actions if isinstance(action, MultiAction) else [action]
		for a in actions:
			effective = TriggerComponent._strip_trigger(a).strip()
			if isinstance(effective, AxisAction) or isinstance(effective, MouseAction):
				if analog:
					# UI can do only one analog action per trigger
					return False, half, full, analog
				analog = a
			elif isinstance(a, TriggerAction):
				# New format:
				#   half pressed: release_level == TRIGGER_MAX (255)
				#   full pressed: release_level == press_level
				# Old format:
				#   half and full both used release_level == TRIGGER_MAX.
				#   A lone action below TRIGGER_CLICK is treated as half;
				#   one at TRIGGER_CLICK is treated as legacy full.
				if full and half:
					# UI can handle only one full and
					# one half-press action
					return False, half, full, analog
				if a.release_level == a.press_level:
					if full:
						if not half and full.release_level == TRIGGER_MAX:
							half, full = full, a
						else:
							return False, half, full, analog
					else:
						full = a
				elif a.release_level == TRIGGER_MAX:
					if full and a.press_level < full.press_level:
						if half:
							# UI can handle only one half-press action
							return False, half, full, analog
						half = a
					elif full:
						if half:
							# UI can handle only one half-press action
							return False, half, full, analog
						half, full = full, a
					else:
						full = a
				else:
					if half:
						# UI can handle only one half-press action
						return False, half, full, analog
					half = a
			elif isinstance(a, HipfireAction):
				hipfire_actions = TriggerComponent._strip_hipfire(a)
				half, full = (x.strip() for x in hipfire_actions)

			elif isinstance(a, NoAction):
				# Ignore theese
				pass
			else:
				# Unhandled action type
				return False, half, full, analog
		# Older profiles used (254, 255) for a full-only action and the
		# same (level, 255) form for a partial-only action.  Preserve the
		# common default-level cases while new profiles use equal levels for
		# an unambiguous full-only action.
		if full and not half and full.press_level < TRIGGER_CLICK:
			full, half = NoAction(), full
		return True, half, full, analog

	@staticmethod
	def _strip_trigger(action):
		"""If passed action is TriggerAction, returns its child action.

		Returns passed action otherwise.
		"""
		if isinstance(action, TriggerAction):
			return action.action
		return action

	@staticmethod
	def _strip_hipfire(action):
		"""If passed action is HipfireAction, returns its childs action.

		Returns passed action otherwise.
		"""
		if isinstance(action, HipfireAction):
			return action.partialpress_action, action.fullpress_action
		return action

	def get_button_title(self):
		return _("Key or Button")

	def set_action(self, mode, action):
		self.half, self.full, self.analog = NoAction(), NoAction(), NoAction()
		sucess, half, full, analog = TriggerComponent._split(action)
		if sucess:
			self._recursing = True
			cb = self.builder.get_object("cbActionType")
			if isinstance(action, HipfireAction):
				self.half, self.full = (TriggerComponent._strip_hipfire(x) for x in (half, full))
				if half and full:
					self.builder.get_object("sclPartialLevel").set_value(action.partialpress_level)
					self.builder.get_object("sclFullLevel").set_value(action.fullpress_level)
					trigger_style = action.mode
					self.set_cb(cb, "HIPFIRE_" + trigger_style, 1)
					self.builder.get_object("sclTimeOut").set_value(action.timeout)

			else:
				self.half, self.full, self.analog = (TriggerComponent._strip_trigger(x) for x in (half, full, analog))
				if half:
					self.builder.get_object("sclPartialLevel").set_value(half.press_level)
					trigger_style = "NORMAL_EXCLUSIVE" if (half.release_level < TRIGGER_MAX) else "NORMAL"
					self.set_cb(cb, trigger_style, 1)
				if full:
					self.builder.get_object("sclFullLevel").set_value(full.press_level)

				if isinstance(analog, AxisAction):
					self.builder.get_object("sclARangeStart").set_value(analog.min)
					self.builder.get_object("sclARangeEnd").set_value(analog.max)

			self._recursing = False
		self.update()

	def update(self):
		self.builder.get_object("lblPartPressed").set_label(describe_action(Action.AC_BUTTON, ButtonAction, self.half))
		self.builder.get_object("lblFullPressed").set_label(describe_action(Action.AC_BUTTON, ButtonAction, self.full))
		self.builder.get_object("lblAnalog").set_label(describe_action(Action.AC_TRIGGER, AxisAction, self.analog))

	def send(self):
		actions = []
		half_level = int(self.builder.get_object("sclPartialLevel").get_value())
		full_level = int(self.builder.get_object("sclFullLevel").get_value())
		cb = self.builder.get_object("cbActionType")
		trigger_style = cb.get_model().get_value(cb.get_active_iter(), 1)
		timeout = self.builder.get_object("sclTimeOut").get_value()

		if (trigger_style == "HIPFIRE_NORMAL") and self.half and self.full:
			actions.append(HipfireAction(half_level, full_level, self.half, self.full, HIPFIRE_NORMAL, timeout))
		elif (trigger_style == "HIPFIRE_EXCLUSIVE") and self.half and self.full:
			actions.append(HipfireAction(half_level, full_level, self.half, self.full, HIPFIRE_EXCLUSIVE, timeout))
		elif (trigger_style == "HIPFIRE_SENSIBLE") and self.half and self.full:
			actions.append(HipfireAction(half_level, full_level, self.half, self.full, HIPFIRE_SENSIBLE, timeout))
		else:
			if self.half:
				if self.full and trigger_style == "NORMAL_EXCLUSIVE":
					actions.append(TriggerAction(half_level, full_level, self.half))
				else:
					actions.append(TriggerAction(half_level, TRIGGER_MAX, self.half))
			if self.full:
				actions.append(TriggerAction(full_level, self.full))

			if self.analog:
				analog_start = int(self.builder.get_object("sclARangeStart").get_value())
				analog_end = int(self.builder.get_object("sclARangeEnd").get_value())
				if analog_start == TRIGGER_MIN and analog_end == TRIGGER_MAX:
					actions.append(AxisAction(self.analog.id))
				else:
					actions.append(AxisAction(self.analog.id, analog_start, analog_end))

		self.editor.set_action(MultiAction.make(*actions))

	def on_btFullPressedClear_clicked(self, *a):
		self.full = NoAction()
		self.update()
		self.send()

	def on_btAnalogClear_clicked(self, *a):
		self.analog = NoAction()
		self.update()
		self.send()

	def on_ui_value_changed(self, *a):
		if not self._recursing:
			self.send()

	def on_cbActionType_changed(self, *a):
		if not self._recursing:
			self.send()

	def on_btPartPressed_clicked(self, *a):
		"""'Partialy Pressed Action' handler"""
		ae = self.choose_editor(self.half, "")
		ae.set_title(_("Select Partialy Pressed Action"))
		ae.hide_name()
		ae.set_input("half", self.half, mode=Action.AC_BUTTON)
		ae.show(self.editor.window)

	def on_btFullPress_clicked(self, *a):
		"""'Fully Pressed Action' handler"""
		ae = self.choose_editor(self.full, "")
		ae.set_title(_("Select Fully Pressed Action"))
		ae.hide_name()
		ae.set_input("full", self.full, mode=Action.AC_BUTTON)
		ae.show(self.editor.window)

	def on_btAnalog_clicked(self, *a):
		"""'Analog Output' handler."""
		b = SimpleChooser(self.app, "axis", lambda action: self.on_action_chosen("analog", action))
		b.set_title(_("Select Analog Axis"))
		b.display_action(Action.AC_STICK, AxisAction(self.analog))
		b.show(self.editor.window)

	def on_action_chosen(self, i, action, mark_changed=True):
		if i == "full":
			self.full = action
		elif i == "half":
			self.half = action
		else:
			self.analog = action
		self.update()
		self.send()

	def on_btFullyPresedClear_clicked(self, *a):
		self.builder.get_object("sclFullLevel").set_value(TRIGGER_CLICK)

	def on_btPartPresedClear_clicked(self, *a):
		self.builder.get_object("sclPartialLevel").set_value(TRIGGER_HALF)

	def on_btARangeStartClear_clicked(self, *a):
		self.builder.get_object("sclARangeStart").set_value(TRIGGER_MIN)

	def on_btARangeEndClear_clicked(self, *a):
		self.builder.get_object("sclARangeEnd").set_value(TRIGGER_MAX)

	def on_btTimeOutClear_clicked(self, *a):
		self.builder.get_object("sclTimeOut").set_value(HipfireAction.DEFAULT_TIMEOUT)



"""Generic linux daemon base class."""

# Adapted from http://www.jejik.com/files/examples/daemon3x.py
# thanks to the original author

import atexit
import os
import signal
import sys
import syslog
import time


class Daemon:
	"""A generic daemon class.

	Usage: subclass the daemon class and override the run() method.
	"""

	def __init__(self, pidfile):
		self.pidfile = pidfile

	def daemonize(self):
		"""Deamonize class. UNIX double fork mechanism."""
		try:
			pid = os.fork()
			if pid > 0:
				# exit first parent
				sys.exit(0)
		except OSError as err:
			sys.stderr.write(f"fork #1 failed: {err}\n")
			sys.exit(1)

		# decouple from parent environment
		os.chdir("/")
		os.setsid()
		os.umask(0)

		# do second fork
		try:
			pid = os.fork()
			if pid > 0:
				# exit from second parent
				sys.exit(0)
		except OSError as err:
			sys.stderr.write(f"fork #2 failed: {err}\n")
			sys.exit(1)

		# redirect standard file descriptors
		sys.stdout.flush()
		sys.stderr.flush()
		stdi = open(os.devnull)
		stdo = open(os.devnull, "a+")
		stde = open(os.devnull, "a+")

		os.dup2(stdi.fileno(), sys.stdin.fileno())
		os.dup2(stdo.fileno(), sys.stdout.fileno())
		os.dup2(stde.fileno(), sys.stderr.fileno())

		# write pidfile
		self.write_pid()

	def write_pid(self):
		"""Write pid file"""
		atexit.register(self.delpid)

		pid = str(os.getpid())
		with open(self.pidfile, "w+") as fd:
			fd.write(pid + "\n")

	def delpid(self):
		"""Delete pid file"""
		os.remove(self.pidfile)

	def start(self, foreground=False):
		"""Start the daemon."""
		# Check for a pidfile to see if the daemon already runs
		try:
			with open(self.pidfile) as pidf:
				pid = int(pidf.read().strip())
		except Exception:
			pid = None

		if pid:
			# Check if PID coresponds to running daemon process and fail if yes
			try:
				assert os.path.exists("/proc")  # Just in case of BSD...
				with open("/proc/%s/cmdline" % (pid,)) as file:
					cmdline = file.read().replace("\x00", " ").strip()
				if sys.argv[0] in cmdline:
					raise Exception("already running")
			except OSError:
				# No such process
				pass
			except:
				message = "pidfile {0} already exist. " + "Daemon already running?\n"
				sys.stderr.write(message.format(self.pidfile))
				sys.exit(1)

			sys.stderr.write("Overwriting stale pidfile\n")

		# Start the daemon
		if not foreground:
			self.daemonize()
		else:
			self.write_pid()
		syslog.syslog(syslog.LOG_INFO, f"{os.path.basename(sys.argv[0])}: started")
		self.on_start()
		while True:
			try:
				self.run()
			except Exception as e:  # pylint: disable=W0703
				syslog.syslog(syslog.LOG_ERR, f"{os.path.basename(sys.argv[0])}: {e!s}")
			time.sleep(2)

	def on_start(self):
		pass

	def stop(self, once=False):
		"""Stop the daemon."""
		# Get the pid from the pidfile
		try:
			with open(self.pidfile) as pidf:
				pid = int(pidf.read().strip())
		except Exception:
			pid = None

		if not pid:
			message = "pidfile {0} does not exist. " + "Daemon not running?\n"
			sys.stderr.write(message.format(self.pidfile))
			return  # not an error in a restart

		# Try killing the daemon process
		try:
			for x in range(10):  # Waits max 1s
				os.kill(pid, signal.SIGTERM)
				if once:
					break
				for x in range(50):
					os.kill(pid, 0)
					time.sleep(0.1)
				time.sleep(0.1)
			os.kill(pid, signal.SIGKILL)
		except OSError as err:
			e = str(err.args)
			if e.find("No such process") > 0:
				if os.path.exists(self.pidfile):
					os.remove(self.pidfile)
			else:
				print(str(err.args))
				sys.exit(1)
		syslog.syslog(syslog.LOG_INFO, f"{os.path.basename(sys.argv[0])}: stopped")

	def restart(self):
		"""Restart the daemon."""
		self.stop()
		time.sleep(2)
		self.start()

	def run(self):
		"""You should override this method when you subclass Daemon.

		It will be called after the process has been daemonized by
		start() or restart().
		"""



"""Generated by scripts/generate-uinput-stubs.py; do not edit."""

from enum import IntEnum

class Keys(IntEnum):
	KEY_RESERVED = 0
	KEY_ESC = 1
	KEY_1 = 2
	KEY_2 = 3
	KEY_3 = 4
	KEY_4 = 5
	KEY_5 = 6
	KEY_6 = 7
	KEY_7 = 8
	KEY_8 = 9
	KEY_9 = 10
	KEY_0 = 11
	KEY_MINUS = 12
	KEY_EQUAL = 13
	KEY_BACKSPACE = 14
	KEY_TAB = 15
	KEY_Q = 16
	KEY_W = 17
	KEY_E = 18
	KEY_R = 19
	KEY_T = 20
	KEY_Y = 21
	KEY_U = 22
	KEY_I = 23
	KEY_O = 24
	KEY_P = 25
	KEY_LEFTBRACE = 26
	KEY_RIGHTBRACE = 27
	KEY_ENTER = 28
	KEY_LEFTCTRL = 29
	KEY_A = 30
	KEY_S = 31
	KEY_D = 32
	KEY_F = 33
	KEY_G = 34
	KEY_H = 35
	KEY_J = 36
	KEY_K = 37
	KEY_L = 38
	KEY_SEMICOLON = 39
	KEY_APOSTROPHE = 40
	KEY_GRAVE = 41
	KEY_LEFTSHIFT = 42
	KEY_BACKSLASH = 43
	KEY_Z = 44
	KEY_X = 45
	KEY_C = 46
	KEY_V = 47
	KEY_B = 48
	KEY_N = 49
	KEY_M = 50
	KEY_COMMA = 51
	KEY_DOT = 52
	KEY_SLASH = 53
	KEY_RIGHTSHIFT = 54
	KEY_KPASTERISK = 55
	KEY_LEFTALT = 56
	KEY_SPACE = 57
	KEY_CAPSLOCK = 58
	KEY_F1 = 59
	KEY_F2 = 60
	KEY_F3 = 61
	KEY_F4 = 62
	KEY_F5 = 63
	KEY_F6 = 64
	KEY_F7 = 65
	KEY_F8 = 66
	KEY_F9 = 67
	KEY_F10 = 68
	KEY_NUMLOCK = 69
	KEY_SCROLLLOCK = 70
	KEY_KP7 = 71
	KEY_KP8 = 72
	KEY_KP9 = 73
	KEY_KPMINUS = 74
	KEY_KP4 = 75
	KEY_KP5 = 76
	KEY_KP6 = 77
	KEY_KPPLUS = 78
	KEY_KP1 = 79
	KEY_KP2 = 80
	KEY_KP3 = 81
	KEY_KP0 = 82
	KEY_KPDOT = 83
	KEY_ZENKAKUHANKAKU = 85
	KEY_102ND = 86
	KEY_F11 = 87
	KEY_F12 = 88
	KEY_RO = 89
	KEY_KATAKANA = 90
	KEY_HIRAGANA = 91
	KEY_HENKAN = 92
	KEY_KATAKANAHIRAGANA = 93
	KEY_MUHENKAN = 94
	KEY_KPJPCOMMA = 95
	KEY_KPENTER = 96
	KEY_RIGHTCTRL = 97
	KEY_KPSLASH = 98
	KEY_SYSRQ = 99
	KEY_RIGHTALT = 100
	KEY_LINEFEED = 101
	KEY_HOME = 102
	KEY_UP = 103
	KEY_PAGEUP = 104
	KEY_LEFT = 105
	KEY_RIGHT = 106
	KEY_END = 107
	KEY_DOWN = 108
	KEY_PAGEDOWN = 109
	KEY_INSERT = 110
	KEY_DELETE = 111
	KEY_MACRO = 112
	KEY_MUTE = 113
	KEY_VOLUMEDOWN = 114
	KEY_VOLUMEUP = 115
	KEY_POWER = 116
	KEY_KPEQUAL = 117
	KEY_KPPLUSMINUS = 118
	KEY_PAUSE = 119
	KEY_SCALE = 120
	KEY_KPCOMMA = 121
	KEY_HANGEUL = 122
	KEY_HANGUEL = 122
	KEY_HANJA = 123
	KEY_YEN = 124
	KEY_LEFTMETA = 125
	KEY_RIGHTMETA = 126
	KEY_COMPOSE = 127
	KEY_STOP = 128
	KEY_AGAIN = 129
	KEY_PROPS = 130
	KEY_UNDO = 131
	KEY_FRONT = 132
	KEY_COPY = 133
	KEY_OPEN = 134
	KEY_PASTE = 135
	KEY_FIND = 136
	KEY_CUT = 137
	KEY_HELP = 138
	KEY_MENU = 139
	KEY_CALC = 140
	KEY_SETUP = 141
	KEY_SLEEP = 142
	KEY_WAKEUP = 143
	KEY_FILE = 144
	KEY_SENDFILE = 145
	KEY_DELETEFILE = 146
	KEY_XFER = 147
	KEY_PROG1 = 148
	KEY_PROG2 = 149
	KEY_WWW = 150
	KEY_MSDOS = 151
	KEY_COFFEE = 152
	KEY_SCREENLOCK = 152
	KEY_ROTATE_DISPLAY = 153
	KEY_DIRECTION = 153
	KEY_CYCLEWINDOWS = 154
	KEY_MAIL = 155
	KEY_BOOKMARKS = 156
	KEY_COMPUTER = 157
	KEY_BACK = 158
	KEY_FORWARD = 159
	KEY_CLOSECD = 160
	KEY_EJECTCD = 161
	KEY_EJECTCLOSECD = 162
	KEY_NEXTSONG = 163
	KEY_PLAYPAUSE = 164
	KEY_PREVIOUSSONG = 165
	KEY_STOPCD = 166
	KEY_RECORD = 167
	KEY_REWIND = 168
	KEY_PHONE = 169
	KEY_ISO = 170
	KEY_CONFIG = 171
	KEY_HOMEPAGE = 172
	KEY_REFRESH = 173
	KEY_EXIT = 174
	KEY_MOVE = 175
	KEY_EDIT = 176
	KEY_SCROLLUP = 177
	KEY_SCROLLDOWN = 178
	KEY_KPLEFTPAREN = 179
	KEY_KPRIGHTPAREN = 180
	KEY_NEW = 181
	KEY_REDO = 182
	KEY_F13 = 183
	KEY_F14 = 184
	KEY_F15 = 185
	KEY_F16 = 186
	KEY_F17 = 187
	KEY_F18 = 188
	KEY_F19 = 189
	KEY_F20 = 190
	KEY_F21 = 191
	KEY_F22 = 192
	KEY_F23 = 193
	KEY_F24 = 194
	KEY_PLAYCD = 200
	KEY_PAUSECD = 201
	KEY_PROG3 = 202
	KEY_PROG4 = 203
	KEY_ALL_APPLICATIONS = 204
	KEY_DASHBOARD = 204
	KEY_SUSPEND = 205
	KEY_CLOSE = 206
	KEY_PLAY = 207
	KEY_FASTFORWARD = 208
	KEY_BASSBOOST = 209
	KEY_PRINT = 210
	KEY_HP = 211
	KEY_CAMERA = 212
	KEY_SOUND = 213
	KEY_QUESTION = 214
	KEY_EMAIL = 215
	KEY_CHAT = 216
	KEY_SEARCH = 217
	KEY_CONNECT = 218
	KEY_FINANCE = 219
	KEY_SPORT = 220
	KEY_SHOP = 221
	KEY_ALTERASE = 222
	KEY_CANCEL = 223
	KEY_BRIGHTNESSDOWN = 224
	KEY_BRIGHTNESSUP = 225
	KEY_MEDIA = 226
	KEY_SWITCHVIDEOMODE = 227
	KEY_KBDILLUMTOGGLE = 228
	KEY_KBDILLUMDOWN = 229
	KEY_KBDILLUMUP = 230
	KEY_SEND = 231
	KEY_REPLY = 232
	KEY_FORWARDMAIL = 233
	KEY_SAVE = 234
	KEY_DOCUMENTS = 235
	KEY_BATTERY = 236
	KEY_BLUETOOTH = 237
	KEY_WLAN = 238
	KEY_UWB = 239
	KEY_UNKNOWN = 240
	KEY_VIDEO_NEXT = 241
	KEY_VIDEO_PREV = 242
	KEY_BRIGHTNESS_CYCLE = 243
	KEY_BRIGHTNESS_AUTO = 244
	KEY_BRIGHTNESS_ZERO = 244
	KEY_DISPLAY_OFF = 245
	KEY_WWAN = 246
	KEY_WIMAX = 246
	KEY_RFKILL = 247
	KEY_MICMUTE = 248
	BTN_MISC = 256
	BTN_0 = 256
	BTN_1 = 257
	BTN_2 = 258
	BTN_3 = 259
	BTN_4 = 260
	BTN_5 = 261
	BTN_6 = 262
	BTN_7 = 263
	BTN_8 = 264
	BTN_9 = 265
	BTN_MOUSE = 272
	BTN_LEFT = 272
	BTN_RIGHT = 273
	BTN_MIDDLE = 274
	BTN_SIDE = 275
	BTN_EXTRA = 276
	BTN_FORWARD = 277
	BTN_BACK = 278
	BTN_TASK = 279
	BTN_JOYSTICK = 288
	BTN_TRIGGER = 288
	BTN_THUMB = 289
	BTN_THUMB2 = 290
	BTN_TOP = 291
	BTN_TOP2 = 292
	BTN_PINKIE = 293
	BTN_BASE = 294
	BTN_BASE2 = 295
	BTN_BASE3 = 296
	BTN_BASE4 = 297
	BTN_BASE5 = 298
	BTN_BASE6 = 299
	BTN_DEAD = 303
	BTN_GAMEPAD = 304
	BTN_SOUTH = 304
	BTN_A = 304
	BTN_EAST = 305
	BTN_B = 305
	BTN_C = 306
	BTN_NORTH = 307
	BTN_X = 307
	BTN_WEST = 308
	BTN_Y = 308
	BTN_Z = 309
	BTN_TL = 310
	BTN_TR = 311
	BTN_TL2 = 312
	BTN_TR2 = 313
	BTN_SELECT = 314
	BTN_START = 315
	BTN_MODE = 316
	BTN_THUMBL = 317
	BTN_THUMBR = 318
	BTN_DIGI = 320
	BTN_TOOL_PEN = 320
	BTN_TOOL_RUBBER = 321
	BTN_TOOL_BRUSH = 322
	BTN_TOOL_PENCIL = 323
	BTN_TOOL_AIRBRUSH = 324
	BTN_TOOL_FINGER = 325
	BTN_TOOL_MOUSE = 326
	BTN_TOOL_LENS = 327
	BTN_TOOL_QUINTTAP = 328
	BTN_STYLUS3 = 329
	BTN_TOUCH = 330
	BTN_STYLUS = 331
	BTN_STYLUS2 = 332
	BTN_TOOL_DOUBLETAP = 333
	BTN_TOOL_TRIPLETAP = 334
	BTN_TOOL_QUADTAP = 335
	BTN_WHEEL = 336
	BTN_GEAR_DOWN = 336
	BTN_GEAR_UP = 337
	KEY_OK = 352
	KEY_SELECT = 353
	KEY_GOTO = 354
	KEY_CLEAR = 355
	KEY_POWER2 = 356
	KEY_OPTION = 357
	KEY_INFO = 358
	KEY_TIME = 359
	KEY_VENDOR = 360
	KEY_ARCHIVE = 361
	KEY_PROGRAM = 362
	KEY_CHANNEL = 363
	KEY_FAVORITES = 364
	KEY_EPG = 365
	KEY_PVR = 366
	KEY_MHP = 367
	KEY_LANGUAGE = 368
	KEY_TITLE = 369
	KEY_SUBTITLE = 370
	KEY_ANGLE = 371
	KEY_FULL_SCREEN = 372
	KEY_ZOOM = 372
	KEY_MODE = 373
	KEY_KEYBOARD = 374
	KEY_ASPECT_RATIO = 375
	KEY_SCREEN = 375
	KEY_PC = 376
	KEY_TV = 377
	KEY_TV2 = 378
	KEY_VCR = 379
	KEY_VCR2 = 380
	KEY_SAT = 381
	KEY_SAT2 = 382
	KEY_CD = 383
	KEY_TAPE = 384
	KEY_RADIO = 385
	KEY_TUNER = 386
	KEY_PLAYER = 387
	KEY_TEXT = 388
	KEY_DVD = 389
	KEY_AUX = 390
	KEY_MP3 = 391
	KEY_AUDIO = 392
	KEY_VIDEO = 393
	KEY_DIRECTORY = 394
	KEY_LIST = 395
	KEY_MEMO = 396
	KEY_CALENDAR = 397
	KEY_RED = 398
	KEY_GREEN = 399
	KEY_YELLOW = 400
	KEY_BLUE = 401
	KEY_CHANNELUP = 402
	KEY_CHANNELDOWN = 403
	KEY_FIRST = 404
	KEY_LAST = 405
	KEY_AB = 406
	KEY_NEXT = 407
	KEY_RESTART = 408
	KEY_SLOW = 409
	KEY_SHUFFLE = 410
	KEY_BREAK = 411
	KEY_PREVIOUS = 412
	KEY_DIGITS = 413
	KEY_TEEN = 414
	KEY_TWEN = 415
	KEY_VIDEOPHONE = 416
	KEY_GAMES = 417
	KEY_ZOOMIN = 418
	KEY_ZOOMOUT = 419
	KEY_ZOOMRESET = 420
	KEY_WORDPROCESSOR = 421
	KEY_EDITOR = 422
	KEY_SPREADSHEET = 423
	KEY_GRAPHICSEDITOR = 424
	KEY_PRESENTATION = 425
	KEY_DATABASE = 426
	KEY_NEWS = 427
	KEY_VOICEMAIL = 428
	KEY_ADDRESSBOOK = 429
	KEY_MESSENGER = 430
	KEY_DISPLAYTOGGLE = 431
	KEY_BRIGHTNESS_TOGGLE = 431
	KEY_SPELLCHECK = 432
	KEY_LOGOFF = 433
	KEY_DOLLAR = 434
	KEY_EURO = 435
	KEY_FRAMEBACK = 436
	KEY_FRAMEFORWARD = 437
	KEY_CONTEXT_MENU = 438
	KEY_MEDIA_REPEAT = 439
	KEY_10CHANNELSUP = 440
	KEY_10CHANNELSDOWN = 441
	KEY_IMAGES = 442
	KEY_NOTIFICATION_CENTER = 444
	KEY_PICKUP_PHONE = 445
	KEY_HANGUP_PHONE = 446
	KEY_LINK_PHONE = 447
	KEY_DEL_EOL = 448
	KEY_DEL_EOS = 449
	KEY_INS_LINE = 450
	KEY_DEL_LINE = 451
	KEY_FN = 464
	KEY_FN_ESC = 465
	KEY_FN_F1 = 466
	KEY_FN_F2 = 467
	KEY_FN_F3 = 468
	KEY_FN_F4 = 469
	KEY_FN_F5 = 470
	KEY_FN_F6 = 471
	KEY_FN_F7 = 472
	KEY_FN_F8 = 473
	KEY_FN_F9 = 474
	KEY_FN_F10 = 475
	KEY_FN_F11 = 476
	KEY_FN_F12 = 477
	KEY_FN_1 = 478
	KEY_FN_2 = 479
	KEY_FN_D = 480
	KEY_FN_E = 481
	KEY_FN_F = 482
	KEY_FN_S = 483
	KEY_FN_B = 484
	KEY_FN_RIGHT_SHIFT = 485
	KEY_BRL_DOT1 = 497
	KEY_BRL_DOT2 = 498
	KEY_BRL_DOT3 = 499
	KEY_BRL_DOT4 = 500
	KEY_BRL_DOT5 = 501
	KEY_BRL_DOT6 = 502
	KEY_BRL_DOT7 = 503
	KEY_BRL_DOT8 = 504
	KEY_BRL_DOT9 = 505
	KEY_BRL_DOT10 = 506
	KEY_NUMERIC_0 = 512
	KEY_NUMERIC_1 = 513
	KEY_NUMERIC_2 = 514
	KEY_NUMERIC_3 = 515
	KEY_NUMERIC_4 = 516
	KEY_NUMERIC_5 = 517
	KEY_NUMERIC_6 = 518
	KEY_NUMERIC_7 = 519
	KEY_NUMERIC_8 = 520
	KEY_NUMERIC_9 = 521
	KEY_NUMERIC_STAR = 522
	KEY_NUMERIC_POUND = 523
	KEY_NUMERIC_A = 524
	KEY_NUMERIC_B = 525
	KEY_NUMERIC_C = 526
	KEY_NUMERIC_D = 527
	KEY_CAMERA_FOCUS = 528
	KEY_WPS_BUTTON = 529
	KEY_TOUCHPAD_TOGGLE = 530
	KEY_TOUCHPAD_ON = 531
	KEY_TOUCHPAD_OFF = 532
	KEY_CAMERA_ZOOMIN = 533
	KEY_CAMERA_ZOOMOUT = 534
	KEY_CAMERA_UP = 535
	KEY_CAMERA_DOWN = 536
	KEY_CAMERA_LEFT = 537
	KEY_CAMERA_RIGHT = 538
	KEY_ATTENDANT_ON = 539
	KEY_ATTENDANT_OFF = 540
	KEY_ATTENDANT_TOGGLE = 541
	KEY_LIGHTS_TOGGLE = 542
	BTN_DPAD_UP = 544
	BTN_DPAD_DOWN = 545
	BTN_DPAD_LEFT = 546
	BTN_DPAD_RIGHT = 547
	BTN_GRIPL = 548
	BTN_GRIPR = 549
	BTN_GRIPL2 = 550
	BTN_GRIPR2 = 551
	KEY_ALS_TOGGLE = 560
	KEY_ROTATE_LOCK_TOGGLE = 561
	KEY_REFRESH_RATE_TOGGLE = 562
	KEY_BUTTONCONFIG = 576
	KEY_TASKMANAGER = 577
	KEY_JOURNAL = 578
	KEY_CONTROLPANEL = 579
	KEY_APPSELECT = 580
	KEY_SCREENSAVER = 581
	KEY_VOICECOMMAND = 582
	KEY_ASSISTANT = 583
	KEY_KBD_LAYOUT_NEXT = 584
	KEY_EMOJI_PICKER = 585
	KEY_DICTATE = 586
	KEY_CAMERA_ACCESS_ENABLE = 587
	KEY_CAMERA_ACCESS_DISABLE = 588
	KEY_CAMERA_ACCESS_TOGGLE = 589
	KEY_ACCESSIBILITY = 590
	KEY_DO_NOT_DISTURB = 591
	KEY_BRIGHTNESS_MIN = 592
	KEY_BRIGHTNESS_MAX = 593
	KEY_EPRIVACY_SCREEN_ON = 594
	KEY_EPRIVACY_SCREEN_OFF = 595
	KEY_ACTION_ON_SELECTION = 596
	KEY_CONTEXTUAL_INSERT = 597
	KEY_CONTEXTUAL_QUERY = 598
	KEY_KBDINPUTASSIST_PREV = 608
	KEY_KBDINPUTASSIST_NEXT = 609
	KEY_KBDINPUTASSIST_PREVGROUP = 610
	KEY_KBDINPUTASSIST_NEXTGROUP = 611
	KEY_KBDINPUTASSIST_ACCEPT = 612
	KEY_KBDINPUTASSIST_CANCEL = 613
	KEY_RIGHT_UP = 614
	KEY_RIGHT_DOWN = 615
	KEY_LEFT_UP = 616
	KEY_LEFT_DOWN = 617
	KEY_ROOT_MENU = 618
	KEY_MEDIA_TOP_MENU = 619
	KEY_NUMERIC_11 = 620
	KEY_NUMERIC_12 = 621
	KEY_AUDIO_DESC = 622
	KEY_3D_MODE = 623
	KEY_NEXT_FAVORITE = 624
	KEY_STOP_RECORD = 625
	KEY_PAUSE_RECORD = 626
	KEY_VOD = 627
	KEY_UNMUTE = 628
	KEY_FASTREVERSE = 629
	KEY_SLOWREVERSE = 630
	KEY_DATA = 631
	KEY_ONSCREEN_KEYBOARD = 632
	KEY_PRIVACY_SCREEN_TOGGLE = 633
	KEY_SELECTIVE_SCREENSHOT = 634
	KEY_NEXT_ELEMENT = 635
	KEY_PREVIOUS_ELEMENT = 636
	KEY_AUTOPILOT_ENGAGE_TOGGLE = 637
	KEY_MARK_WAYPOINT = 638
	KEY_SOS = 639
	KEY_NAV_CHART = 640
	KEY_FISHING_CHART = 641
	KEY_SINGLE_RANGE_RADAR = 642
	KEY_DUAL_RANGE_RADAR = 643
	KEY_RADAR_OVERLAY = 644
	KEY_TRADITIONAL_SONAR = 645
	KEY_CLEARVU_SONAR = 646
	KEY_SIDEVU_SONAR = 647
	KEY_NAV_INFO = 648
	KEY_BRIGHTNESS_MENU = 649
	KEY_MACRO1 = 656
	KEY_MACRO2 = 657
	KEY_MACRO3 = 658
	KEY_MACRO4 = 659
	KEY_MACRO5 = 660
	KEY_MACRO6 = 661
	KEY_MACRO7 = 662
	KEY_MACRO8 = 663
	KEY_MACRO9 = 664
	KEY_MACRO10 = 665
	KEY_MACRO11 = 666
	KEY_MACRO12 = 667
	KEY_MACRO13 = 668
	KEY_MACRO14 = 669
	KEY_MACRO15 = 670
	KEY_MACRO16 = 671
	KEY_MACRO17 = 672
	KEY_MACRO18 = 673
	KEY_MACRO19 = 674
	KEY_MACRO20 = 675
	KEY_MACRO21 = 676
	KEY_MACRO22 = 677
	KEY_MACRO23 = 678
	KEY_MACRO24 = 679
	KEY_MACRO25 = 680
	KEY_MACRO26 = 681
	KEY_MACRO27 = 682
	KEY_MACRO28 = 683
	KEY_MACRO29 = 684
	KEY_MACRO30 = 685
	KEY_MACRO_RECORD_START = 688
	KEY_MACRO_RECORD_STOP = 689
	KEY_MACRO_PRESET_CYCLE = 690
	KEY_MACRO_PRESET1 = 691
	KEY_MACRO_PRESET2 = 692
	KEY_MACRO_PRESET3 = 693
	KEY_KBD_LCD_MENU1 = 696
	KEY_KBD_LCD_MENU2 = 697
	KEY_KBD_LCD_MENU3 = 698
	KEY_KBD_LCD_MENU4 = 699
	KEY_KBD_LCD_MENU5 = 700
	KEY_PERFORMANCE = 701
	BTN_TRIGGER_HAPPY = 704
	BTN_TRIGGER_HAPPY1 = 704
	BTN_TRIGGER_HAPPY2 = 705
	BTN_TRIGGER_HAPPY3 = 706
	BTN_TRIGGER_HAPPY4 = 707
	BTN_TRIGGER_HAPPY5 = 708
	BTN_TRIGGER_HAPPY6 = 709
	BTN_TRIGGER_HAPPY7 = 710
	BTN_TRIGGER_HAPPY8 = 711
	BTN_TRIGGER_HAPPY9 = 712
	BTN_TRIGGER_HAPPY10 = 713
	BTN_TRIGGER_HAPPY11 = 714
	BTN_TRIGGER_HAPPY12 = 715
	BTN_TRIGGER_HAPPY13 = 716
	BTN_TRIGGER_HAPPY14 = 717
	BTN_TRIGGER_HAPPY15 = 718
	BTN_TRIGGER_HAPPY16 = 719
	BTN_TRIGGER_HAPPY17 = 720
	BTN_TRIGGER_HAPPY18 = 721
	BTN_TRIGGER_HAPPY19 = 722
	BTN_TRIGGER_HAPPY20 = 723
	BTN_TRIGGER_HAPPY21 = 724
	BTN_TRIGGER_HAPPY22 = 725
	BTN_TRIGGER_HAPPY23 = 726
	BTN_TRIGGER_HAPPY24 = 727
	BTN_TRIGGER_HAPPY25 = 728
	BTN_TRIGGER_HAPPY26 = 729
	BTN_TRIGGER_HAPPY27 = 730
	BTN_TRIGGER_HAPPY28 = 731
	BTN_TRIGGER_HAPPY29 = 732
	BTN_TRIGGER_HAPPY30 = 733
	BTN_TRIGGER_HAPPY31 = 734
	BTN_TRIGGER_HAPPY32 = 735
	BTN_TRIGGER_HAPPY33 = 736
	BTN_TRIGGER_HAPPY34 = 737
	BTN_TRIGGER_HAPPY35 = 738
	BTN_TRIGGER_HAPPY36 = 739
	BTN_TRIGGER_HAPPY37 = 740
	BTN_TRIGGER_HAPPY38 = 741
	BTN_TRIGGER_HAPPY39 = 742
	BTN_TRIGGER_HAPPY40 = 743
	KEY_MIN_INTERESTING = 113
	KEY_MAX = 767
	KEY_CNT = 768


class KeysOnly(IntEnum):
	KEY_RESERVED = 0
	KEY_ESC = 1
	KEY_1 = 2
	KEY_2 = 3
	KEY_3 = 4
	KEY_4 = 5
	KEY_5 = 6
	KEY_6 = 7
	KEY_7 = 8
	KEY_8 = 9
	KEY_9 = 10
	KEY_0 = 11
	KEY_MINUS = 12
	KEY_EQUAL = 13
	KEY_BACKSPACE = 14
	KEY_TAB = 15
	KEY_Q = 16
	KEY_W = 17
	KEY_E = 18
	KEY_R = 19
	KEY_T = 20
	KEY_Y = 21
	KEY_U = 22
	KEY_I = 23
	KEY_O = 24
	KEY_P = 25
	KEY_LEFTBRACE = 26
	KEY_RIGHTBRACE = 27
	KEY_ENTER = 28
	KEY_LEFTCTRL = 29
	KEY_A = 30
	KEY_S = 31
	KEY_D = 32
	KEY_F = 33
	KEY_G = 34
	KEY_H = 35
	KEY_J = 36
	KEY_K = 37
	KEY_L = 38
	KEY_SEMICOLON = 39
	KEY_APOSTROPHE = 40
	KEY_GRAVE = 41
	KEY_LEFTSHIFT = 42
	KEY_BACKSLASH = 43
	KEY_Z = 44
	KEY_X = 45
	KEY_C = 46
	KEY_V = 47
	KEY_B = 48
	KEY_N = 49
	KEY_M = 50
	KEY_COMMA = 51
	KEY_DOT = 52
	KEY_SLASH = 53
	KEY_RIGHTSHIFT = 54
	KEY_KPASTERISK = 55
	KEY_LEFTALT = 56
	KEY_SPACE = 57
	KEY_CAPSLOCK = 58
	KEY_F1 = 59
	KEY_F2 = 60
	KEY_F3 = 61
	KEY_F4 = 62
	KEY_F5 = 63
	KEY_F6 = 64
	KEY_F7 = 65
	KEY_F8 = 66
	KEY_F9 = 67
	KEY_F10 = 68
	KEY_NUMLOCK = 69
	KEY_SCROLLLOCK = 70
	KEY_KP7 = 71
	KEY_KP8 = 72
	KEY_KP9 = 73
	KEY_KPMINUS = 74
	KEY_KP4 = 75
	KEY_KP5 = 76
	KEY_KP6 = 77
	KEY_KPPLUS = 78
	KEY_KP1 = 79
	KEY_KP2 = 80
	KEY_KP3 = 81
	KEY_KP0 = 82
	KEY_KPDOT = 83
	KEY_ZENKAKUHANKAKU = 85
	KEY_102ND = 86
	KEY_F11 = 87
	KEY_F12 = 88
	KEY_RO = 89
	KEY_KATAKANA = 90
	KEY_HIRAGANA = 91
	KEY_HENKAN = 92
	KEY_KATAKANAHIRAGANA = 93
	KEY_MUHENKAN = 94
	KEY_KPJPCOMMA = 95
	KEY_KPENTER = 96
	KEY_RIGHTCTRL = 97
	KEY_KPSLASH = 98
	KEY_SYSRQ = 99
	KEY_RIGHTALT = 100
	KEY_LINEFEED = 101
	KEY_HOME = 102
	KEY_UP = 103
	KEY_PAGEUP = 104
	KEY_LEFT = 105
	KEY_RIGHT = 106
	KEY_END = 107
	KEY_DOWN = 108
	KEY_PAGEDOWN = 109
	KEY_INSERT = 110
	KEY_DELETE = 111
	KEY_MACRO = 112
	KEY_MUTE = 113
	KEY_VOLUMEDOWN = 114
	KEY_VOLUMEUP = 115
	KEY_POWER = 116
	KEY_KPEQUAL = 117
	KEY_KPPLUSMINUS = 118
	KEY_PAUSE = 119
	KEY_SCALE = 120
	KEY_KPCOMMA = 121
	KEY_HANGEUL = 122
	KEY_HANGUEL = 122
	KEY_HANJA = 123
	KEY_YEN = 124
	KEY_LEFTMETA = 125
	KEY_RIGHTMETA = 126
	KEY_COMPOSE = 127
	KEY_STOP = 128
	KEY_AGAIN = 129
	KEY_PROPS = 130
	KEY_UNDO = 131
	KEY_FRONT = 132
	KEY_COPY = 133
	KEY_OPEN = 134
	KEY_PASTE = 135
	KEY_FIND = 136
	KEY_CUT = 137
	KEY_HELP = 138
	KEY_MENU = 139
	KEY_CALC = 140
	KEY_SETUP = 141
	KEY_SLEEP = 142
	KEY_WAKEUP = 143
	KEY_FILE = 144
	KEY_SENDFILE = 145
	KEY_DELETEFILE = 146
	KEY_XFER = 147
	KEY_PROG1 = 148
	KEY_PROG2 = 149
	KEY_WWW = 150
	KEY_MSDOS = 151
	KEY_COFFEE = 152
	KEY_SCREENLOCK = 152
	KEY_ROTATE_DISPLAY = 153
	KEY_DIRECTION = 153
	KEY_CYCLEWINDOWS = 154
	KEY_MAIL = 155
	KEY_BOOKMARKS = 156
	KEY_COMPUTER = 157
	KEY_BACK = 158
	KEY_FORWARD = 159
	KEY_CLOSECD = 160
	KEY_EJECTCD = 161
	KEY_EJECTCLOSECD = 162
	KEY_NEXTSONG = 163
	KEY_PLAYPAUSE = 164
	KEY_PREVIOUSSONG = 165
	KEY_STOPCD = 166
	KEY_RECORD = 167
	KEY_REWIND = 168
	KEY_PHONE = 169
	KEY_ISO = 170
	KEY_CONFIG = 171
	KEY_HOMEPAGE = 172
	KEY_REFRESH = 173
	KEY_EXIT = 174
	KEY_MOVE = 175
	KEY_EDIT = 176
	KEY_SCROLLUP = 177
	KEY_SCROLLDOWN = 178
	KEY_KPLEFTPAREN = 179
	KEY_KPRIGHTPAREN = 180
	KEY_NEW = 181
	KEY_REDO = 182
	KEY_F13 = 183
	KEY_F14 = 184
	KEY_F15 = 185
	KEY_F16 = 186
	KEY_F17 = 187
	KEY_F18 = 188
	KEY_F19 = 189
	KEY_F20 = 190
	KEY_F21 = 191
	KEY_F22 = 192
	KEY_F23 = 193
	KEY_F24 = 194
	KEY_PLAYCD = 200
	KEY_PAUSECD = 201
	KEY_PROG3 = 202
	KEY_PROG4 = 203
	KEY_ALL_APPLICATIONS = 204
	KEY_DASHBOARD = 204
	KEY_SUSPEND = 205
	KEY_CLOSE = 206
	KEY_PLAY = 207
	KEY_FASTFORWARD = 208
	KEY_BASSBOOST = 209
	KEY_PRINT = 210
	KEY_HP = 211
	KEY_CAMERA = 212
	KEY_SOUND = 213
	KEY_QUESTION = 214
	KEY_EMAIL = 215
	KEY_CHAT = 216
	KEY_SEARCH = 217
	KEY_CONNECT = 218
	KEY_FINANCE = 219
	KEY_SPORT = 220
	KEY_SHOP = 221
	KEY_ALTERASE = 222
	KEY_CANCEL = 223
	KEY_BRIGHTNESSDOWN = 224
	KEY_BRIGHTNESSUP = 225
	KEY_MEDIA = 226
	KEY_SWITCHVIDEOMODE = 227
	KEY_KBDILLUMTOGGLE = 228
	KEY_KBDILLUMDOWN = 229
	KEY_KBDILLUMUP = 230
	KEY_SEND = 231
	KEY_REPLY = 232
	KEY_FORWARDMAIL = 233
	KEY_SAVE = 234
	KEY_DOCUMENTS = 235
	KEY_BATTERY = 236
	KEY_BLUETOOTH = 237
	KEY_WLAN = 238
	KEY_UWB = 239
	KEY_UNKNOWN = 240
	KEY_VIDEO_NEXT = 241
	KEY_VIDEO_PREV = 242
	KEY_BRIGHTNESS_CYCLE = 243
	KEY_BRIGHTNESS_AUTO = 244
	KEY_BRIGHTNESS_ZERO = 244
	KEY_DISPLAY_OFF = 245
	KEY_WWAN = 246
	KEY_WIMAX = 246
	KEY_RFKILL = 247
	KEY_MICMUTE = 248
	KEY_OK = 352
	KEY_SELECT = 353
	KEY_GOTO = 354
	KEY_CLEAR = 355
	KEY_POWER2 = 356
	KEY_OPTION = 357
	KEY_INFO = 358
	KEY_TIME = 359
	KEY_VENDOR = 360
	KEY_ARCHIVE = 361
	KEY_PROGRAM = 362
	KEY_CHANNEL = 363
	KEY_FAVORITES = 364
	KEY_EPG = 365
	KEY_PVR = 366
	KEY_MHP = 367
	KEY_LANGUAGE = 368
	KEY_TITLE = 369
	KEY_SUBTITLE = 370
	KEY_ANGLE = 371
	KEY_FULL_SCREEN = 372
	KEY_ZOOM = 372
	KEY_MODE = 373
	KEY_KEYBOARD = 374
	KEY_ASPECT_RATIO = 375
	KEY_SCREEN = 375
	KEY_PC = 376
	KEY_TV = 377
	KEY_TV2 = 378
	KEY_VCR = 379
	KEY_VCR2 = 380
	KEY_SAT = 381
	KEY_SAT2 = 382
	KEY_CD = 383
	KEY_TAPE = 384
	KEY_RADIO = 385
	KEY_TUNER = 386
	KEY_PLAYER = 387
	KEY_TEXT = 388
	KEY_DVD = 389
	KEY_AUX = 390
	KEY_MP3 = 391
	KEY_AUDIO = 392
	KEY_VIDEO = 393
	KEY_DIRECTORY = 394
	KEY_LIST = 395
	KEY_MEMO = 396
	KEY_CALENDAR = 397
	KEY_RED = 398
	KEY_GREEN = 399
	KEY_YELLOW = 400
	KEY_BLUE = 401
	KEY_CHANNELUP = 402
	KEY_CHANNELDOWN = 403
	KEY_FIRST = 404
	KEY_LAST = 405
	KEY_AB = 406
	KEY_NEXT = 407
	KEY_RESTART = 408
	KEY_SLOW = 409
	KEY_SHUFFLE = 410
	KEY_BREAK = 411
	KEY_PREVIOUS = 412
	KEY_DIGITS = 413
	KEY_TEEN = 414
	KEY_TWEN = 415
	KEY_VIDEOPHONE = 416
	KEY_GAMES = 417
	KEY_ZOOMIN = 418
	KEY_ZOOMOUT = 419
	KEY_ZOOMRESET = 420
	KEY_WORDPROCESSOR = 421
	KEY_EDITOR = 422
	KEY_SPREADSHEET = 423
	KEY_GRAPHICSEDITOR = 424
	KEY_PRESENTATION = 425
	KEY_DATABASE = 426
	KEY_NEWS = 427
	KEY_VOICEMAIL = 428
	KEY_ADDRESSBOOK = 429
	KEY_MESSENGER = 430
	KEY_DISPLAYTOGGLE = 431
	KEY_BRIGHTNESS_TOGGLE = 431
	KEY_SPELLCHECK = 432
	KEY_LOGOFF = 433
	KEY_DOLLAR = 434
	KEY_EURO = 435
	KEY_FRAMEBACK = 436
	KEY_FRAMEFORWARD = 437
	KEY_CONTEXT_MENU = 438
	KEY_MEDIA_REPEAT = 439
	KEY_10CHANNELSUP = 440
	KEY_10CHANNELSDOWN = 441
	KEY_IMAGES = 442
	KEY_NOTIFICATION_CENTER = 444
	KEY_PICKUP_PHONE = 445
	KEY_HANGUP_PHONE = 446
	KEY_LINK_PHONE = 447
	KEY_DEL_EOL = 448
	KEY_DEL_EOS = 449
	KEY_INS_LINE = 450
	KEY_DEL_LINE = 451
	KEY_FN = 464
	KEY_FN_ESC = 465
	KEY_FN_F1 = 466
	KEY_FN_F2 = 467
	KEY_FN_F3 = 468
	KEY_FN_F4 = 469
	KEY_FN_F5 = 470
	KEY_FN_F6 = 471
	KEY_FN_F7 = 472
	KEY_FN_F8 = 473
	KEY_FN_F9 = 474
	KEY_FN_F10 = 475
	KEY_FN_F11 = 476
	KEY_FN_F12 = 477
	KEY_FN_1 = 478
	KEY_FN_2 = 479
	KEY_FN_D = 480
	KEY_FN_E = 481
	KEY_FN_F = 482
	KEY_FN_S = 483
	KEY_FN_B = 484
	KEY_FN_RIGHT_SHIFT = 485
	KEY_BRL_DOT1 = 497
	KEY_BRL_DOT2 = 498
	KEY_BRL_DOT3 = 499
	KEY_BRL_DOT4 = 500
	KEY_BRL_DOT5 = 501
	KEY_BRL_DOT6 = 502
	KEY_BRL_DOT7 = 503
	KEY_BRL_DOT8 = 504
	KEY_BRL_DOT9 = 505
	KEY_BRL_DOT10 = 506
	KEY_NUMERIC_0 = 512
	KEY_NUMERIC_1 = 513
	KEY_NUMERIC_2 = 514
	KEY_NUMERIC_3 = 515
	KEY_NUMERIC_4 = 516
	KEY_NUMERIC_5 = 517
	KEY_NUMERIC_6 = 518
	KEY_NUMERIC_7 = 519
	KEY_NUMERIC_8 = 520
	KEY_NUMERIC_9 = 521
	KEY_NUMERIC_STAR = 522
	KEY_NUMERIC_POUND = 523
	KEY_NUMERIC_A = 524
	KEY_NUMERIC_B = 525
	KEY_NUMERIC_C = 526
	KEY_NUMERIC_D = 527
	KEY_CAMERA_FOCUS = 528
	KEY_WPS_BUTTON = 529
	KEY_TOUCHPAD_TOGGLE = 530
	KEY_TOUCHPAD_ON = 531
	KEY_TOUCHPAD_OFF = 532
	KEY_CAMERA_ZOOMIN = 533
	KEY_CAMERA_ZOOMOUT = 534
	KEY_CAMERA_UP = 535
	KEY_CAMERA_DOWN = 536
	KEY_CAMERA_LEFT = 537
	KEY_CAMERA_RIGHT = 538
	KEY_ATTENDANT_ON = 539
	KEY_ATTENDANT_OFF = 540
	KEY_ATTENDANT_TOGGLE = 541
	KEY_LIGHTS_TOGGLE = 542
	KEY_ALS_TOGGLE = 560
	KEY_ROTATE_LOCK_TOGGLE = 561
	KEY_REFRESH_RATE_TOGGLE = 562
	KEY_BUTTONCONFIG = 576
	KEY_TASKMANAGER = 577
	KEY_JOURNAL = 578
	KEY_CONTROLPANEL = 579
	KEY_APPSELECT = 580
	KEY_SCREENSAVER = 581
	KEY_VOICECOMMAND = 582
	KEY_ASSISTANT = 583
	KEY_KBD_LAYOUT_NEXT = 584
	KEY_EMOJI_PICKER = 585
	KEY_DICTATE = 586
	KEY_CAMERA_ACCESS_ENABLE = 587
	KEY_CAMERA_ACCESS_DISABLE = 588
	KEY_CAMERA_ACCESS_TOGGLE = 589
	KEY_ACCESSIBILITY = 590
	KEY_DO_NOT_DISTURB = 591
	KEY_BRIGHTNESS_MIN = 592
	KEY_BRIGHTNESS_MAX = 593
	KEY_EPRIVACY_SCREEN_ON = 594
	KEY_EPRIVACY_SCREEN_OFF = 595
	KEY_ACTION_ON_SELECTION = 596
	KEY_CONTEXTUAL_INSERT = 597
	KEY_CONTEXTUAL_QUERY = 598
	KEY_KBDINPUTASSIST_PREV = 608
	KEY_KBDINPUTASSIST_NEXT = 609
	KEY_KBDINPUTASSIST_PREVGROUP = 610
	KEY_KBDINPUTASSIST_NEXTGROUP = 611
	KEY_KBDINPUTASSIST_ACCEPT = 612
	KEY_KBDINPUTASSIST_CANCEL = 613
	KEY_RIGHT_UP = 614
	KEY_RIGHT_DOWN = 615
	KEY_LEFT_UP = 616
	KEY_LEFT_DOWN = 617
	KEY_ROOT_MENU = 618
	KEY_MEDIA_TOP_MENU = 619
	KEY_NUMERIC_11 = 620
	KEY_NUMERIC_12 = 621
	KEY_AUDIO_DESC = 622
	KEY_3D_MODE = 623
	KEY_NEXT_FAVORITE = 624
	KEY_STOP_RECORD = 625
	KEY_PAUSE_RECORD = 626
	KEY_VOD = 627
	KEY_UNMUTE = 628
	KEY_FASTREVERSE = 629
	KEY_SLOWREVERSE = 630
	KEY_DATA = 631
	KEY_ONSCREEN_KEYBOARD = 632
	KEY_PRIVACY_SCREEN_TOGGLE = 633
	KEY_SELECTIVE_SCREENSHOT = 634
	KEY_NEXT_ELEMENT = 635
	KEY_PREVIOUS_ELEMENT = 636
	KEY_AUTOPILOT_ENGAGE_TOGGLE = 637
	KEY_MARK_WAYPOINT = 638
	KEY_SOS = 639
	KEY_NAV_CHART = 640
	KEY_FISHING_CHART = 641
	KEY_SINGLE_RANGE_RADAR = 642
	KEY_DUAL_RANGE_RADAR = 643
	KEY_RADAR_OVERLAY = 644
	KEY_TRADITIONAL_SONAR = 645
	KEY_CLEARVU_SONAR = 646
	KEY_SIDEVU_SONAR = 647
	KEY_NAV_INFO = 648
	KEY_BRIGHTNESS_MENU = 649
	KEY_MACRO1 = 656
	KEY_MACRO2 = 657
	KEY_MACRO3 = 658
	KEY_MACRO4 = 659
	KEY_MACRO5 = 660
	KEY_MACRO6 = 661
	KEY_MACRO7 = 662
	KEY_MACRO8 = 663
	KEY_MACRO9 = 664
	KEY_MACRO10 = 665
	KEY_MACRO11 = 666
	KEY_MACRO12 = 667
	KEY_MACRO13 = 668
	KEY_MACRO14 = 669
	KEY_MACRO15 = 670
	KEY_MACRO16 = 671
	KEY_MACRO17 = 672
	KEY_MACRO18 = 673
	KEY_MACRO19 = 674
	KEY_MACRO20 = 675
	KEY_MACRO21 = 676
	KEY_MACRO22 = 677
	KEY_MACRO23 = 678
	KEY_MACRO24 = 679
	KEY_MACRO25 = 680
	KEY_MACRO26 = 681
	KEY_MACRO27 = 682
	KEY_MACRO28 = 683
	KEY_MACRO29 = 684
	KEY_MACRO30 = 685
	KEY_MACRO_RECORD_START = 688
	KEY_MACRO_RECORD_STOP = 689
	KEY_MACRO_PRESET_CYCLE = 690
	KEY_MACRO_PRESET1 = 691
	KEY_MACRO_PRESET2 = 692
	KEY_MACRO_PRESET3 = 693
	KEY_KBD_LCD_MENU1 = 696
	KEY_KBD_LCD_MENU2 = 697
	KEY_KBD_LCD_MENU3 = 698
	KEY_KBD_LCD_MENU4 = 699
	KEY_KBD_LCD_MENU5 = 700
	KEY_PERFORMANCE = 701
	KEY_MIN_INTERESTING = 113
	KEY_MAX = 767
	KEY_CNT = 768


class Axes(IntEnum):
	ABS_X = 0
	ABS_Y = 1
	ABS_Z = 2
	ABS_RX = 3
	ABS_RY = 4
	ABS_RZ = 5
	ABS_THROTTLE = 6
	ABS_RUDDER = 7
	ABS_WHEEL = 8
	ABS_GAS = 9
	ABS_BRAKE = 10
	ABS_HAT0X = 16
	ABS_HAT0Y = 17
	ABS_HAT1X = 18
	ABS_HAT1Y = 19
	ABS_HAT2X = 20
	ABS_HAT2Y = 21
	ABS_HAT3X = 22
	ABS_HAT3Y = 23
	ABS_PRESSURE = 24
	ABS_DISTANCE = 25
	ABS_TILT_X = 26
	ABS_TILT_Y = 27
	ABS_TOOL_WIDTH = 28
	ABS_VOLUME = 32
	ABS_PROFILE = 33
	ABS_SND_PROFILE = 34
	ABS_MISC = 40
	ABS_RESERVED = 46
	ABS_MT_SLOT = 47
	ABS_MT_TOUCH_MAJOR = 48
	ABS_MT_TOUCH_MINOR = 49
	ABS_MT_WIDTH_MAJOR = 50
	ABS_MT_WIDTH_MINOR = 51
	ABS_MT_ORIENTATION = 52
	ABS_MT_POSITION_X = 53
	ABS_MT_POSITION_Y = 54
	ABS_MT_TOOL_TYPE = 55
	ABS_MT_BLOB_ID = 56
	ABS_MT_TRACKING_ID = 57
	ABS_MT_PRESSURE = 58
	ABS_MT_DISTANCE = 59
	ABS_MT_TOOL_X = 60
	ABS_MT_TOOL_Y = 61
	ABS_MAX = 63
	ABS_CNT = 64


class Rels(IntEnum):
	REL_X = 0
	REL_Y = 1
	REL_Z = 2
	REL_RX = 3
	REL_RY = 4
	REL_RZ = 5
	REL_HWHEEL = 6
	REL_DIAL = 7
	REL_WHEEL = 8
	REL_MISC = 9
	REL_RESERVED = 10
	REL_WHEEL_HI_RES = 11
	REL_HWHEEL_HI_RES = 12
	REL_MAX = 15
	REL_CNT = 16



#!/usr/bin/env python3
import signal
import sys


def sigint(*a):
	print("\n*break*")
	sys.exit(-1)


def main():
	signal.signal(signal.SIGINT, sigint)

	import gi

	gi.require_version("Gtk", "3.0")
	gi.require_version("Rsvg", "2.0")
	gi.require_version("GdkX11", "3.0")

	from scc.tools import init_logging

	init_logging()

	from scc.osd.binding_display import BindingDisplay

	d = BindingDisplay()
	if not d.parse_argumets(sys.argv):
		sys.exit(1)
	d.run()
	sys.exit(d.get_exit_code())


if __name__ == "__main__":
	main()



"""SC Controller - Menu Data

Container for list of menu items + required parsers
"""

import json
import os

from scc.actions import Action
from scc.tools import _


class MenuData:
	"""Contains list of menu items. Indexable"""

	def __init__(self, *items):
		self.__items = list(items)

	def generate(self, menuhandler):
		"""Converts all generators into MenuItems (by calling .generate() on them)
		and returns generated MenuData.

		Returns new MenuData instance.
		"""
		items = []
		for i in self:
			if isinstance(i, MenuGenerator):
				items.extend(i.generate(menuhandler))
			else:
				items.append(i)
		return MenuData(*items)

	def compress(self):
		for i in self.__items:
			if i.action:
				i.action = i.action.compress()

	def __len__(self):
		return len(self.__items)

	def __getitem__(self, index):
		return self.__items[index]

	def __iter__(self):
		return iter(self.__items)

	def get_all_actions(self):
		"""Returns generator with every action defined in this menu, including
		child actions.
		"""
		for item in self:
			if hasattr(item, "action") and item.action:
				for i in item.action.get_all_actions():
					yield i

	def get_by_id(self, id):
		"""Returns item with specified ID.
		Throws KeyError if there is no such item.
		"""
		for a in self:
			if a.id == id:
				return a
		raise KeyError("No such item")

	def index(self, a):
		return self.__items.index(a)

	def encode(self):
		"""Returns menu data as dict storable in json (profile) file"""
		rv = []
		for i in self:
			rv.append(i.encode())
		return rv

	@staticmethod
	def from_args(data):
		"""Parses list of arguments in [id1, label1, id2, label2 ...] format.
		Throws ValueError if number of items in 'data' is odd.
		"""
		if len(data) % 2 != 0:
			raise ValueError("Odd number of items")
		if len(data) < 1:
			raise ValueError("Not items")

		# Rearange data into list of pair tuples
		data = [(data[i * 2], data[(i * 2) + 1]) for i in range(len(data) / 2)]

		# Parse data
		m = MenuData()
		for id, label in data:
			m.__items.append(MenuItem(id, label))
		return m

	@staticmethod
	def from_json_data(data, action_parser=None):
		"""Loads menu from parsed JSON dict.
		Actions are parsed only if action_parser is set to ActionParser instance.
		"""
		m = MenuData()
		used_ids = set()
		for i in data:
			item = None
			if "generator" in i and i["generator"] in MENU_GENERATORS:
				item = MENU_GENERATORS[i["generator"]](**i)
			elif "separator" in i:
				item = Separator(i["name"] if "name" in i else None)
			elif "submenu" in i:
				item = Submenu(
					i["submenu"],
					i["name"] if "name" in i else None,
					icon=i["icon"] if "icon" in i else None,
				)
			elif "id" not in i:
				# Cannot add menu without ID
				continue
			else:
				action = None
				id = i["id"]
				if id in used_ids:
					# Cannot add duplicate ID
					continue
				if action_parser:
					action = action_parser.from_json_data(i)
				used_ids.add(id)
				label, icon = id, None
				if "name" in i:
					label = i["name"]
				elif action:
					label = action.describe(Action.AC_OSD)
				if "icon" in i:
					icon = i["icon"]
				item = MenuItem(id, label, action, icon=icon)
			m.__items.append(item)

		return m

	@staticmethod
	def from_fileobj(fileobj, action_parser=None):
		"""Loads menu from file-like object.
		Actions are parsed only if action_parser is set to ActionParser instance.
		"""
		data = json.loads(fileobj.read())
		return MenuData.from_json_data(data, action_parser)

	@staticmethod
	def from_file(filename, action_parser=None):
		"""Loads menu from file.
		Actions are parsed only if action_parser is set to ActionParser instance.
		"""
		return MenuData.from_fileobj(open(filename), action_parser)

	@staticmethod
	def from_profile(filename, menuname, action_parser=None):
		"""Loads menu from JSON profile file.
		Actions are parsed only if action_parser is set to ActionParser instance.

		Menus are stored as list under <root>/menus/<menuname>.
		Throws ValueError if specified file cannot be parsed or
		specified menu cannot be found.
		"""
		data = json.loads(open(filename).read())
		if "menus" not in data:
			raise ValueError("Menu not found")
		if menuname not in data["menus"]:
			raise ValueError("Menu not found")

		return MenuData.from_json_data(data["menus"][menuname], action_parser)


class MenuItem:
	"""Really just dummy container"""

	def __init__(self, id, label, action=None, callback=None, icon=None):
		self.id = id
		self.label = label
		self.action = action
		self.icon = icon
		self.callback = callback  # If defined, called when user chooses menu instead of using action
		self.widget = None  # May be set by UI code

	def describe(self):
		"""Returns user-friendly description of MenuItem or MenuGenerator.
		"""
		return self.label

	def encode(self):
		"""Returns item data as dict storable in json (profile) file"""
		if self.action and type(self.action) in (str,):
			rv = {"action": self.action}
		elif self.action:
			rv = self.action.encode()
		else:
			rv = {}
		rv["id"] = self.id
		rv["name"] = self.label
		if self.icon:
			rv["icon"] = self.icon
		return rv


class Separator(MenuItem):
	"""Internally, separator is MenuItem without action and id"""

	def __init__(self, label=None):
		MenuItem.__init__(self, None, label)

	def describe(self):
		if self.label:
			return _("----[ %s ]----") % (self.label,)
		return _("---- Separator ----")

	def encode(self):
		if self.label:
			return {"separator": True, "name": self.label}
		return {"separator": True}


class Submenu(MenuItem):
	"""Internally, separator is MenuItem without action and id"""

	def __init__(self, filename, label=None, icon=None):
		if not label:
			label = ".".join(os.path.split(filename)[-1].split(".")[0:-1])
		self.filename = filename
		MenuItem.__init__(self, str(id(self)), label=label, icon=icon)

	def describe(self):
		return self.label + "  " + _(">>")

	def encode(self):
		rv = {"submenu": self.filename}
		if self.label:
			rv["name"] = self.label
		if self.icon:
			rv["icon"] = self.icon
		return rv


class MenuGenerator:
	GENERATOR_NAME = None
	""" Generates list of MenuItems """

	def __init__(self, **b):
		"""Passed are all keys loaded from json dict that defined this generator.
		__init__ of generator should ignore all unknown keys.
		"""
		self.id = None  # Used only in editor
		self.icon = None  # same

	def describe(self):
		"""Returns user-friendly description of MenuItem or MenuGenerator.
		"""
		return "[ %s ] " % (self.__class__.__name__,)

	def encode(self):
		"""Returns generator data as dict storable in json (profile) file"""
		return {"generator": self.GENERATOR_NAME}

	def generate(self, menuhandler):
		return []


# Holds dict of knowm menu ganerators, but generated elsewhere
MENU_GENERATORS = {}



"""SC Controller - Action Editor

Allows to edit button or trigger action.
"""

import logging
import os

from gi.repository import Gdk, Gtk

log = logging.getLogger("Editor")


class ComboSetter:
	def set_cb(self, cb, key, keyindex=0) -> bool:
		"""Set combobox value.

		Returns True on success or False if key is not found.
		"""
		model = cb.get_model()
		self._recursing = True
		for row in model:
			if key == row[keyindex]:
				cb.set_active_iter(row.iter)
				self._recursing = False
				return True
		log.warning("Failed to set combobox value, key '%s' not found", key)
		self._recursing = False
		return False


class Editor(ComboSetter):
	"""Common stuff for all editor windows"""

	ERROR_CSS = " #error {background-color:green; color:red;} "
	_error_css_provider = None

	def __init__(self):
		self.added_widget = None  # See add_widget method

	def on_window_key_press_event(self, trash, event):
		"""Checks if pressed key was escape and if yes, closes window"""
		if event.keyval == Gdk.KEY_Escape:
			self.close()

	def setup_widgets(self):
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(self.app.gladepath, self.GLADE))
		self.window = self.builder.get_object("Dialog")
		self.builder.connect_signals(self)

	@staticmethod
	def install_error_css():
		if Editor._error_css_provider is None:
			Editor._error_css_provider = Gtk.CssProvider()
			Editor._error_css_provider.load_from_data(Editor.ERROR_CSS.encode("utf-8"))
			Gtk.StyleContext.add_provider_for_screen(
				Gdk.Screen.get_default(), Editor._error_css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER,
			)

	def hide_dont_destroy(self, w, *a):
		"""When used as handler for 'delete-event' signal, prevents window from being destroyed after closing."""
		w.hide()
		return True

	def set_title(self, title):
		self.window.set_title(title)
		self.builder.get_object("header").set_title(title)

	def close(self, *a):
		self.window.destroy()

	def get_transient_for(self):
		"""Return parent window for this editor. Usually main application window"""
		return self._transient_for

	def show(self, transient_for):
		if transient_for:
			self._transient_for = transient_for
			self.window.set_transient_for(transient_for)
			self.window.set_modal(True)
		self.window.show()

	def add_widget(self, label, widget):
		"""Add new widget into row before Action Name.

		Widget is automatically passed to Macro Editor or Modeshift Editor
		if either one is opened from editor window.

		When editor window is closed or destroyed, widget is automatically
		deattached to keep it from destroying.
		"""
		lblAddedWidget = self.builder.get_object("lblAddedWidget")
		vbAddedWidget = self.builder.get_object("vbAddedWidget")
		lblAddedWidget.set_label(label)
		lblAddedWidget.set_visible(True)
		for ch in vbAddedWidget.get_children():
			vbAddedWidget.remove(ch)
		self.added_widget = widget
		vbAddedWidget.pack_start(widget, True, False, 0)
		vbAddedWidget.set_visible(True)

	def remove_added_widget(self):
		"""Remove added widget, if any.

		Should be called from on_destory handlers.
		"""
		vbAddedWidget = self.builder.get_object("vbAddedWidget")
		for ch in vbAddedWidget.get_children():
			vbAddedWidget.remove(ch)
		self.added_widget = None

	def send_added_widget(self, target):
		"""Transfer added widget to new editor window"""
		if self.added_widget:
			vbAddedWidget = self.builder.get_object("vbAddedWidget")
			lblAddedWidget = self.builder.get_object("lblAddedWidget")
			label = lblAddedWidget.get_label()
			w = self.added_widget
			self.remove_added_widget()
			target.add_widget(label, w)



# The MIT License (MIT)
#
# Copyright (c) 2015 Stany MARCEL <stanypub@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from enum import IntEnum
from importlib.metadata import packages_distributions, version

distribution_name = packages_distributions()[__package__][0]

"""
If SC Controller is updated while daemon is running, DAEMON_VERSION send by
daemon will differ one one expected by UI and daemon will be forcefully restarted.
"""
DAEMON_VERSION = version(distribution_name)

HPERIOD  = 0.02
LPERIOD  = 0.5
DURATION = 1.0

# Constants used when forcing gamepad to read some type of event is needed
FE_STICK   = 1
FE_TRIGGER = 2
FE_PAD     = 3
FE_GYRO    = 4

# Trigger names, pads, etc. These constants are used on multiple places
LEFT   = "LEFT"
RIGHT  = "RIGHT"
CPAD   = "CPAD"
RSTICK = "RSTICK"
DPAD   = "DPAD"
WHOLE  = "WHOLE"
STICK  = "STICK"
GYRO   = "GYRO"
PITCH  = "PITCH"
YAW    = "YAW"
ROLL   = "ROLL"

# Special constants currently used only by menus
SAME    = "SAME"    # Menu is canceled by releasing same button that intiated it
DEFAULT = "DEFAULT" # Default confirm/cancel button. A/B for menus initiated by
                    # button, pad clicking / releasing for menus on pads

# Deadzone modes
CUT     = "CUT"
ROUND   = "ROUND"
LINEAR  = "LINEAR"
MINIMUM = "MINIMUM"

# Hipfire modes
HIPFIRE_NORMAL    = "NORMAL"
HIPFIRE_SENSIBLE  = "SENSIBLE"
HIPFIRE_EXCLUSIVE = "EXCLUSIVE"

PARSER_CONSTANTS = ( LEFT, RIGHT, WHOLE, STICK, GYRO, PITCH,
	YAW, ROLL, DEFAULT, SAME, CUT, ROUND, LINEAR, MINIMUM,
	HIPFIRE_NORMAL, HIPFIRE_SENSIBLE, HIPFIRE_EXCLUSIVE )



class SCButtons(IntEnum):
	RPADTOUCH   = 0b000010000000000000000000000000000
	LPADTOUCH   = 0b000001000000000000000000000000000
	RPAD        = 0b000000100000000000000000000000000
	LPAD        = 0b000000010000000000000000000000000 # Same for stick but without LPadTouch
	RGRIP       = 0b000000001000000000000000000000000
	LGRIP       = 0b000000000100000000000000000000000
	START       = 0b000000000010000000000000000000000
	C           = 0b000000000001000000000000000000000
	BACK        = 0b000000000000100000000000000000000
	A           = 0b000000000000000001000000000000000
	X           = 0b000000000000000000100000000000000
	B           = 0b000000000000000000010000000000000
	Y           = 0b000000000000000000001000000000000
	LB          = 0b000000000000000000000100000000000
	RB          = 0b000000000000000000000010000000000
	LT          = 0b000000000000000000000001000000000
	RT          = 0b000000000000000000000000100000000
	CPADTOUCH   = 0b000000000000000000000000000000100 # Available on DS4 pad
	CPADPRESS   = 0b000000000000000000000000000000010 # Available on DS4 pad
	STICKPRESS  = 0b001000000000000000000000000000000
	RSTICKPRESS = 0b010000000000000000000000000000000
	DOTS        = 0b000000000000000000000000000001000 # Deck only
	RGRIP2      = 0b000000000000000000000000000100000 # Deck only
	LGRIP2      = 0b000000000000000000000000000010000 # Deck only


# If lpad and stick is used at once, this is sent as
# button with every other packet to signalize that
# value of lpad_x and lpad_y belongs to stick
STICKTILT = 0b10000000000000000000000000000000


class HapticPos(IntEnum):
	"""Specify witch pad or trig is used."""

	RIGHT = 0
	LEFT  = 1
	BOTH  = 2 # emulated


class ControllerFlags(IntEnum):
	"""Used by mapper to workaround some physical differences between Steam Controller and other pads."""

	NONE           =      0 # No flags, default SC.
	HAS_RSTICK     = 1 << 0 # Controller has right stick instead of touchpad
	SEPARATE_STICK = 1 << 1 # Left stick and left pad are using separate axes
	EUREL_GYROS    = 1 << 2 # Gyro sensor values are provided as pitch, yaw
	                        # and roll instead of quaterion. 'q4' is unused
	                        # in such case.
	HAS_CPAD       = 1 << 3 # Controller has DS4-like touchpad in center
	HAS_DPAD       = 1 << 4 # Controller has normal d-pad instead of touchpad
	NO_GRIPS       = 1 << 5 # Controller has no grips
	IS_DECK        = 1 << 6 # Very special case


STICK_PAD_MIN      = -32768
STICK_PAD_MAX      = 32767
STICK_PAD_MIN_HALF = STICK_PAD_MIN / 3
STICK_PAD_MAX_HALF = STICK_PAD_MAX / 3
STICK_PAD_RES      = STICK_PAD_MAX - (STICK_PAD_MIN)

# Take async 360 stick axes into account
OUTPUT_360_STICK_MAX = 32767
OUTPUT_360_STICK_MIN = -32768
OUTPUT_360_STICK_RES = OUTPUT_360_STICK_MAX - (OUTPUT_360_STICK_MIN)

CPAD_MIN   = 0
CPAD_X_MAX = 1916
CPAD_Y_MAX = 930

STICK_PAD_MIN_HALF     = STICK_PAD_MIN / 3
TRIGGER_MIN            = 0
TRIGGER_HALF           = 50
TRIGGER_CLICK          = 254 # Values under this are generated until trigger clicks
TRIGGER_MAX            = 255
BASE_STICK_MOUSE_SPEED = 1000 # Pixels per second



"""SC-Controller - DaemonManager

Starts, kills and controls sccdaemon instance.

I'd call it DaemonController normally, but having something with
full name of "Steam Controller Controller Daemon Controller" sounds
probably too crazy even for me.
"""

import json
import logging
import os

from gi.repository import Gio, GLib, GObject

from scc.constants import SCButtons
from scc.gui import BUTTON_ORDER
from scc.paths import get_daemon_socket
from scc.tools import find_binary, find_button_image, nameof

log = logging.getLogger("DaemonCtrl")


class DaemonManager(GObject.GObject):
	"""Communicates with daemon socket and provides wrappers around everything
	it can do.

	List of signals:
		alive ()
			Emited after daemon is started or found to be alraedy running

		controller-count-changed(count)
			Emited after daemon reports change in controller count, ie when
			new controller is connnected or disconnected.
			Also emited shortly after connection to daemon is initiated.

		dead ()
			Emited after daemon is killed (or exits for some other reason)

		error (description)
			Emited when daemon reports error, most likely not being able to
			access to USB dongle.

		event (controller, pad_stick_or_button, values)
			As 'event' signal on Controller. Allows for capturing events from
			all controllers using single signal.

		profile-changed (profile)
			Emited after profile set for first controller is changed.
			Profile is filename of currently active profile

		reconfigured()
			Emited when daemon reports change in configuration file

		unknown-msg (message)
			Emited when message that can't be parsed internally
			is received from daemon.

		version (ver)
			Emited daemon reports its version - usually only once per connection.
	"""

	__gsignals__ = {
		"alive": (GObject.SignalFlags.RUN_FIRST, None, ()),
		"controller-count-changed": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
		"dead": (GObject.SignalFlags.RUN_FIRST, None, ()),
		"error": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
		"event": (GObject.SignalFlags.RUN_FIRST, None, (object, object, object)),
		"profile-changed": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
		"reconfigured": (GObject.SignalFlags.RUN_FIRST, None, ()),
		"unknown-msg": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
		"version": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
	}

	RECONNECT_INTERVAL = 5

	def __init__(self):
		GObject.GObject.__init__(self)
		self.alive = None
		self.connection = None
		self.connecting = False
		self.buffer = b""
		self._connect()
		self._requests = []
		self._controllers = []  # Ordered as daemon says
		self._controller_by_id = {}  # Source of memory leak

	def get_controllers(self):
		"""Returns list of all controllers connected to daemon.
		Value is cached locally.
		"""
		return [] + self._controllers

	def get_controller(self, controller_id, type=None):
		"""Returns ControllerManager instance bound to provided controller_id.
		Note that this method will return instance for any controller_id,
		even if controller with such ID is not connected to daemon.

		For same controller_id, there is always same instance returned.
		"""
		if controller_id not in self._controller_by_id:
			c = self._controller_by_id[controller_id] = ControllerManager(self, controller_id, type)
		else:
			c = self._controller_by_id[controller_id]
			c.set_type(type)
		return c

	def has_controller(self):
		"""Returns True if there is at lease one controller connected to daemon.
		"""
		return len(self._controllers) > 0

	def _connect(self):
		if self.connecting:
			return
		self.connecting = True
		sc = Gio.SocketClient()
		address = Gio.UnixSocketAddress.new(get_daemon_socket())
		sc.connect_async(address, None, self._on_connected)

	def _on_daemon_died(self, *a):
		"""Called from various places when daemon looks like dead"""
		# Log stuff
		if self.alive is True:
			log.debug("Connection to daemon lost")
		if self.alive is True or self.alive is None:
			self.alive = False
			self.emit("dead")
		self.alive = False
		# Close connection, if any
		if self.connection is not None:
			self.connection.close()
			self.connection = None
		# Emit event
		# Try to reconnect
		GLib.timeout_add_seconds(self.RECONNECT_INTERVAL, self._connect)

	def _on_connected(self, sc, results):
		"""Called when connection to daemon socket is initiated"""
		self.connecting = False
		try:
			self.connection = sc.connect_finish(results)
			if self.connection == None:
				raise Exception("Unknown error")
		except Exception:
			self._on_daemon_died()
			return
		self.buffer = b""
		self.connection.get_input_stream().read_bytes_async(102400, 1, None, self._on_read_data)

	def _on_read_data(self, sc, results):
		"""Called when daemon sends some data"""
		try:
			response = sc.read_bytes_finish(results)
			if response == None:
				raise Exception("No data received")
		except Exception:
			# Broken sonnection, daemon was probbaly terminated
			self._on_daemon_died()
			return
		data = response.get_data()
		if len(data) == 0:
			# Connection terminated
			self._on_daemon_died()
			return
		self.buffer += data
		while b"\n" in self.buffer:
			line, self.buffer = self.buffer.split(b"\n", 1)
			line = line.decode("utf-8")
			if line.startswith("Version:"):
				version = line.split(":", 1)[-1].strip()
				log.debug("Connected to daemon, version %s", version)
				self.emit("version", version)
			elif line.startswith("Ready."):
				log.debug("Daemon is ready.")
				self.alive = True
				self.emit("alive")
			elif line.startswith("OK."):
				if len(self._requests) > 0:
					success_cb, error_cb = self._requests[0]
					self._requests = self._requests[1:]
					success_cb()
			elif line.startswith("Fail:"):
				if len(self._requests) > 0:
					success_cb, error_cb = self._requests[0]
					self._requests = self._requests[1:]
					error_cb(line[5:].strip())
			elif line.startswith("Controller:"):
				controller_id, type, flags, config_file = line[11:].strip().split(" ", 3)
				c = self.get_controller(controller_id, type)
				c._connected = True
				c._type = type
				c._flags = int(flags)
				c._config_file = None if config_file in ("", "None") else config_file
				while c in self._controllers:
					self._controllers.remove(c)
				self._controllers.append(c)
			elif line.startswith("Controller profile:"):
				controller_id, profile = line[19:].strip().split(" ", 1)
				c = self.get_controller(controller_id)
				c._profile = profile.strip()
				c.emit("profile-changed", c._profile)
				log.debug("Daemon reported profile change for %s: %s", controller_id, c._profile)
			elif line.startswith("Controller Count:"):
				count = int(line[17:])
				if count == 0:
					old, self._controllers = self._controllers, []
				else:
					old, self._controllers = self._controllers, self._controllers[-count:]
				self.emit("controller-count-changed", count)
				for c in old:
					if c not in self._controllers:
						c.emit("lost")
			elif line.startswith("Event:"):
				data = line[6:].strip().split(" ")
				c = self.get_controller(data[0])
				c.emit("event", data[1], [int(float(x)) for x in data[2:]])
				self.emit("event", c, data[1], [int(float(x)) for x in data[2:]])
			elif line.startswith("Error:"):
				error = line.split(":", 1)[-1].strip()
				self.alive = True
				log.debug("Daemon reported error '%s'", error)
				self.emit("error", error)
			elif line.startswith("Current profile:"):
				self._profile = line.split(":", 1)[-1].strip()
				self.emit("profile-changed", self._profile)
			elif line.startswith("Reconfigured."):
				self.emit("reconfigured")
			elif line.startswith("PID:") or line == "SCCDaemon":
				# ignore
				pass
			else:
				self.emit("unknown-msg", line)
		# Connection is held forever to detect when daemon exits
		if self.connection:
			self.connection.get_input_stream().read_bytes_async(102400, 1, None, self._on_read_data)

	def is_alive(self):
		"""Returns True if daemon is running"""
		return self.alive

	def request(self, message, success_cb, error_cb):
		"""Creates request and remembers callback for next 'Ok' or 'Fail' message.
		"""
		if self.alive and self.connection is not None:
			tmp = message if type(message) == bytes else bytes(message, "utf-8")
			self._requests.append((success_cb, error_cb))
			(self.connection.get_output_stream().write_all(tmp + b"\n", None))
		else:
			# Instant failure
			error_cb("Not connected.")

	@classmethod
	def nocallback(*a):
		"""Used when request doesn't needs callback"""

	def set_profile(self, filename):
		"""Asks daemon to change 1st controller profile"""
		self.request("Controller.\nProfile: %s" % (filename,), DaemonManager.nocallback, DaemonManager.nocallback)

	def reconfigure(self):
		"""Asks daemon reload configuration file"""
		self.request("Reconfigure.", DaemonManager.nocallback, DaemonManager.nocallback)

	def rescan(self):
		"""Asks daemon to rescan for new devices"""
		self.request("Rescan.", DaemonManager.nocallback, DaemonManager.nocallback)

	def stop(self):
		"""Stops the daemon"""
		Gio.Subprocess.new([find_binary("scc-daemon"), "/dev/null", "stop"], Gio.SubprocessFlags.NONE)
		self.connecting = False

	def start(self, mode="start"):
		"""Starts the daemon and forces connection to be created immediately.
		"""
		if self.alive:
			# Just to clean up living connection
			self.alive = None
			self._on_daemon_died()
		Gio.Subprocess.new([find_binary("scc-daemon"), "/dev/null", mode], Gio.SubprocessFlags.NONE)
		self._connect()
		GLib.timeout_add_seconds(10, self._check_connected)

	def restart(self):
		"""Restarts the daemon and forces connection to be created immediately.
		"""
		self.start(mode="restart")

	def _check_connected(self):
		if not self.alive:
			log.debug("Started daemon but connection is not ready")
			self.emit("error", "CANT_SUMMON_THE_DAEMON")
		return False


class ControllerManager(GObject.GObject):
	"""Represents controller connected to daemon.
	Returned by DaemonManager.get_controller or DaemonManager.get_controllers.

	List of signals:
		event (pad_stick_or_button, values)
			Emited when pad, stick or button is locked using lock() method
			and position or pressed state of that button is changed

		lost()
			Emited when controller is disconnected or turned off

		profile-changed (profile)
			Emited after profile for controller is changed.
			Profile is filename of currently active profile
	"""

	__gsignals__ = {
		"event": (GObject.SignalFlags.RUN_FIRST, None, (object, object)),
		"lost": (GObject.SignalFlags.RUN_FIRST, None, ()),
		"profile-changed": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
	}

	DEFAULT_ICONS = [
		"A",
		"B",
		"X",
		"Y",
		"BACK",
		"C",
		"START",
		"LB",
		"RB",
		"LT",
		"RT",
		"STICK",
		"LPAD",
		"RPAD",
		"RGRIP",
		"LGRIP",
		"DOTS",
	]
	# ^^ those are icon names

	def __init__(self, daemon_manager, controller_id, controller_type):
		GObject.GObject.__init__(self)
		self._dm = daemon_manager
		self._controller_id = controller_id
		self._type = controller_type
		self._config_file = None
		self._connected = False
		self._profile = None
		self._type = None
		self._flags = 0

	def __repr__(self):
		return "<ControllerManager for ID '%s'>" % (self._controller_id,)

	def _send_id(self):
		"""Sends Controller: message to daemon, so next message goes to correct
		controller.
		"""
		self._dm.request("Controller: %s" % (self._controller_id,), DaemonManager.nocallback, DaemonManager.nocallback)

	def is_connected(self):
		"""Returns True, if controller is still connected to daemon.
		Value is cached locally.
		"""
		return self._connected

	def get_type(self):
		"""Returns string identifier of controller driver.

		Value is cached locally, but may be None before controller is connected.
		"""
		return self._type

	def set_type(self, type):
		"""Sets type, if none is yet set.
		"""
		if self._type is None:
			self._type = type

	def get_flags(self):
		"""Returns flags for this controller. See ControllerFlags enum for more info.

		Value is cached locally and returns 0 until controller is connected.
		"""
		return self._flags

	def get_id(self):
		"""Returns ID of this controller. Value is cached locally."""
		return self._controller_id

	def get_gui_config_file(self):
		"""Returns file name of json file that GUI can use to load more data about
		controller (background image, button images, available buttons and
		axes, etc...) File name may be absolute path or just name of file in
		/usr/share/scc

		Returns None if there is no configuration file (GUI will use
		defaults in such case)
		"""
		return self._config_file

	def load_gui_config(self, default_path):
		"""As get_gui_config_file, but returns loaded and parsed config.
		Returns None if config cannot be loaded.
		"""
		filename = self.get_gui_config_file()
		if filename:
			if "/" not in filename:
				filename = os.path.join(default_path, filename)
			try:
				data = json.loads(open(filename).read()) or None
				return data
			except Exception as e:
				log.exception(e)
		return None

	@staticmethod
	def get_button_icon(config, button, prefer_bw=False):
		"""For config returned by load_gui_config() and SCButton constant,
		returns icon filename assigned to that button in controller config or
		default if config is invalid or button unassigned.
		"""
		return find_button_image(ControllerManager.get_button_name(config, button), prefer_bw=prefer_bw)

	@staticmethod
	def get_button_name(config, button):
		"""As get_button_icon, but returns icon name instead of filename.
		"""
		name = nameof(button)
		index = -1
		try:
			if type(button) is str:
				button = SCButtons.__members__[button]
			index = BUTTON_ORDER.index(button)
			name = ControllerManager.DEFAULT_ICONS[index]
			btn_index = config["gui"]["buttons"].index(name)
			if btn_index != -1:
				name = config["gui"]["buttons"][btn_index]
		except:
			pass
		return name

	def get_profile(self):
		"""Returns profile set for this controller. Value is cached locally."""
		return self._profile

	def lock(self, success_cb, error_cb, *what_to_lock):
		"""Locks physical button, axis or pad. Events from locked sources are
		sent to this client and processed using 'event' singal, until
		unlock_all() is called.

		Calls success_cb() on success or error_cb(error) on failure.
		"""
		what = " ".join(what_to_lock)
		self._send_id()
		self._dm.request("Lock: %s" % (what,), success_cb, error_cb)

	def set_led_level(self, value):
		"""Sets brightness of controller led.
		"""
		self._send_id()
		self._dm.request("Led: %s" % (int(value),), DaemonManager.nocallback, DaemonManager.nocallback)

	def set_profile(self, filename):
		"""Asks daemon to change this controller profile"""
		self._send_id()
		self._dm.request("Profile: %s" % (filename,), DaemonManager.nocallback, DaemonManager.nocallback)

	def turnoff(self):
		"""Asks daemon to turn off this controller"""
		self._send_id()
		self._dm.request("Turnoff.", DaemonManager.nocallback, DaemonManager.nocallback)

	def feedback(self, position, amplitude):
		"""Generates feedback effect on controller"""
		self._send_id()
		self._dm.request("Feedback: %s %s" % (position, amplitude), DaemonManager.nocallback, DaemonManager.nocallback)

	def observe(self, success_cb, error_cb, *what_to_lock):
		"""Enables observing on physical button, axis or pad.
		Events from observed sources are sent to this client and processed
		using 'event' singal, until unlock_all() is called.

		Calls success_cb() on success or error_cb(error) on failure.
		"""
		what = " ".join(what_to_lock)
		self._send_id()
		self._dm.request("Observe: %s" % (what,), success_cb, error_cb)

	def replace(self, success_cb, error_cb, what, action):
		"""Temporally replaces action on physical button, axis or pad,
		until unlock_all() is called.

		Calls success_cb() on success or error_cb(error) on failure.
		"""
		actionstr = action.to_string().replace("\n", " ")
		self._dm.request("Replace: %s %s" % (what, actionstr), success_cb, error_cb)

	def unlock_all(self):
		if self._dm.alive:
			self._send_id()
			self._dm.request("Unlock.", lambda *a: False, lambda *a: False)



#!/usr/bin/env python3
"""SC-Controller - Action Editor - On Screen Keyboard Action Component

Assigns actions from scc.osd.osk_actions
"""

import logging

from scc.actions import Action, ButtonAction, NoAction
from scc.constants import LEFT
from scc.gui.ae import AEComponent
from scc.gui.parser import GuiActionParser
from scc.osd.osk_actions import CloseOSKAction, MoveOSKAction, OSKAction, OSKCursorAction, OSKPressAction
from scc.tools import _
from scc.uinput import Keys

log = logging.getLogger("AE.SA")

__all__ = ["OSKActionComponent"]


class OSKActionComponent(AEComponent):
	GLADE = "ae/osk_action.glade"
	NAME = "osk_action"
	CTXS = Action.AC_OSK
	PRIORITY = 2

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		self._recursing = False

	def set_action(self, mode, action):
		cb = self.builder.get_object("cbActionType")
		if isinstance(action, CloseOSKAction):
			self.set_cb(cb, "OSK.close()")
		elif isinstance(action, OSKCursorAction) and action.side == LEFT:
			self.set_cb(cb, "OSK.cursor(LEFT)")
		elif isinstance(action, OSKCursorAction):  # and action.side == RIGHT:
			self.set_cb(cb, "OSK.cursor(RIGHT)")
		elif isinstance(action, OSKPressAction) and action.side == LEFT:
			self.set_cb(cb, "OSK.press(LEFT)")
		elif isinstance(action, OSKPressAction):  # and action.side == RIGHT:
			self.set_cb(cb, "OSK.press(RIGHT)")
		elif isinstance(action, MoveOSKAction):
			self.set_cb(cb, "OSK.move()")
		if isinstance(action, ButtonAction):
			if action.button == Keys.BTN_LEFT:
				self.set_cb(cb, "button(Keys.BTN_LEFT)")
			elif action.button == Keys.BTN_RIGHT:
				self.set_cb(cb, "button(Keys.BTN_RIGHT)")
		else:
			self.set_cb(cb, "None")

	def get_button_title(self):
		return _("On-Screen Keyboard")

	def handles(self, mode, action):
		if isinstance(action, ButtonAction):
			return action.button in (Keys.BTN_LEFT, Keys.BTN_RIGHT)
		return isinstance(action, (NoAction, OSKAction, OSKCursorAction))

	def on_cbActionType_changed(self, *a):
		cbActionType = self.builder.get_object("cbActionType")
		key = cbActionType.get_model().get_value(cbActionType.get_active_iter(), 0)
		self.editor.set_action(GuiActionParser().restart(key).parse())



#!/usr/bin/env python3
"""SC-Controller - Grid OSD Menu

Works as OSD menu, but displays item in (as rectangluar as possible - and
that's usually not very much) grid.
"""


import logging
import math

from gi.repository import Gtk

from scc.menu_data import Separator, Submenu
from scc.osd.menu import Menu, MenuIcon
from scc.tools import find_icon

log = logging.getLogger("osd.gridmenu")


class GridMenu(Menu):
	PREFER_BW_ICONS = True

	def __init__(self, cls="osd-menu"):
		Menu.__init__(self, cls)
		self.ipr = 1  # items per row

	def create_parent(self):
		g = Gtk.Grid()
		g.set_name("osd-menu")
		return g

	def pack_items(self, parent, items):
		if self._size > 0:
			self.ipr = self._size
		else:
			self.ipr = int(math.sqrt(max(1, len(items) - 1)) + 1)
			if len(items) == 6:
				self.ipr = 3  # Special (common) cases
			if len(items) == 8:
				self.ipr = 4  # Special (common) cases
		x, y = 0, 0
		for item in items:
			parent.attach(item.widget, x, y, 1, 1)
			x += 1
			if x >= self.ipr:
				x = 0
				y += 1

	def on_stick_direction(self, trash, x, y):
		if x != 0:
			self.next_item(-x)
		elif y != 0:
			for i in range(self.ipr):
				self.next_item(y)

	def generate_widget(self, item):
		if isinstance(item, Separator):
			# Ignored here
			return None
		if item.id is None:
			# Dummies are ignored as well
			return None
		icon_file, has_colors = find_icon(item.icon, False)
		if icon_file:
			# Gridmenu hides label when icon is displayed
			widget = Gtk.Button()
			widget.set_relief(Gtk.ReliefStyle.NONE)
			widget.set_name("osd-menu-item-big-icon")
			if isinstance(item, Submenu):
				item.callback = self.show_submenu
			icon = MenuIcon(icon_file, has_colors)
			widget.add(icon)
			return widget
		return Menu.generate_widget(self, item)
